import torch
from enum import Enum
from typing import Any, List, TypeVar, Union, Generic, Optional, Tuple
from vidrial.jit.static.shape import Shape
from vidrial.jit.static.stride import Stride
from vidrial.jit.static.int import BaseType, Int, _1, _8, _32, _64, _128, _256

T = TypeVar('T')


class CudaDType(Enum):
    """Enum representing CUDA data types."""
    UINT8 = "uint8_t"
    FLOAT = "float"
    FLOAT32 = "float"
    DOUBLE = "double"
    FLOAT64 = "double"
    INT32 = "int32_t"
    INT64 = "int64_t"
    FLOAT16 = "half_t"
    BFLOAT16 = "bfloat16_t"

    def to_cpp(self) -> str:
        return self.value


def last(x: Union[Stride, Shape]) -> Int:
    """Get the last element of a Stride or Shape."""
    if isinstance(x.elements[-1], Int):
        return x.elements[-1]
    else:
        return last(x.elements[-1])


def compact_major(shape: Shape, left: bool = True, current: Int = _1) -> Stride:
    """Create a compact major Stride from a Shape."""
    elements = shape.elements if left else shape.elements[::-1]
    strides = []
    for elem in elements:
        if isinstance(elem, Int):
            strides.append(current)
            current *= elem
        elif isinstance(elem, Shape):
            stride = compact_major(elem, left, current)
            strides.append(stride)
            current = current * elem.size()
    return Stride(*strides) if left else Stride(*reversed(strides))


class Layout(BaseType):
    """Represents a layout type combining shape and stride."""
    def __init__(self, shape: Shape, stride: Optional[Stride] = None):
        self.shape = shape
        if stride is None:
            stride = compact_major(shape)
        self.stride = stride

    def to_tuple(self) -> Tuple[Tuple[Any, ...], Tuple[Any, ...]]:
        return (self.shape.to_tuple(), self.stride.to_tuple())
            
    def to_cpp(self) -> str:
        return f"Layout<{self.shape.to_cpp()}, {self.stride.to_cpp()}>"

    def __str__(self) -> str:
        return f"{self.shape}:{self.stride}"
    
    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Layout):
            return self.shape == other.shape and self.stride == other.stride
        return False

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(shape={self.shape}, stride={self.stride})"
    
    def __reduce__(self):
        """Make Layout picklable by specifying how to reconstruct it."""
        return (type(self), (self.shape, self.stride))


class Trait(BaseType):
    """Base class for traits with name and dtype."""
    def __init__(self, name: str, dtype: CudaDType):
        self.name = name
        self.dtype = dtype

    def to_cpp(self) -> str:
        return f"{self.name}<{self.dtype.value}>"


class UniversalCopy(Trait):
    """Universal copy trait."""
    def __init__(self, dtype: CudaDType):
        super().__init__("UniversalCopy", dtype)


class CopyAtom(BaseType):
    """Represents a copy atom with trait and dtype."""
    def __init__(self, trait: Trait):
        self.trait = trait

    def to_cpp(self) -> str:
        return f"Copy_Atom<{self.trait.to_cpp()}, {self.trait.dtype.value}>"


ATOMS_A_F16_B_F16 = [
    "SM80_16x8x8_F32F16F16F32_TN",
    "SM80_16x8x16_F32F16F16F32_TN",
]

ATOMS_A_BF16_B_BF16 = [
    "SM80_16x8x8_F32BF16BF16F32_TN",
    "SM80_16x8x16_F32BF16BF16F32_TN"
]

ATOMS_A_TF32_B_TF32 = [
    "SM80_16x8x4_F32TF32TF32F32_TN",
    "SM80_16x8x8_F32TF32TF32F32_TN",
]

def get_valid_atoms(A: torch.Tensor, B: torch.Tensor) -> List[str]:
    """Get the valid atoms for a given pair of tensors."""
    if A.dtype == torch.float16 and B.dtype == torch.float16:
        return ATOMS_A_F16_B_F16
    elif A.dtype == torch.bfloat16 and B.dtype == torch.bfloat16:
        return ATOMS_A_BF16_B_BF16
    elif A.dtype == torch.float32 and B.dtype == torch.float32:
        return ATOMS_A_TF32_B_TF32
    else:
        raise ValueError(f"No valid atoms found for {A.dtype} and {B.dtype}")


# Common trait instances
UniversalCopyFloat = UniversalCopy(CudaDType.FLOAT)
UniversalCopyUint8 = UniversalCopy(CudaDType.UINT8)




from math import prod
import torch as th
from copy import copy
import inspect
import logging
from dataclasses import dataclass
from typing import Tuple, List, Dict, Any, Iterable
from vidrial.py_utils.gpu import get_cuda_device_basic_props
from vidrial.jit.static.types import CudaDType
from vidrial.jit.static.shape import static_shape, Shape
from vidrial.jit.static.util import torch_dtype_to_c
from itertools import product
from functools import lru_cache, partial, wraps

device_props = get_cuda_device_basic_props()[0]
SMEM_LIMIT = device_props['CU_DEVICE_ATTRIBUTE_MAX_SHARED_MEMORY_PER_MULTIPROCESSOR']
REG_PER_SM = device_props['CU_DEVICE_ATTRIBUTE_MAX_REGISTERS_PER_MULTIPROCESSOR']
SM_NUMBER = device_props['CU_DEVICE_ATTRIBUTE_MULTIPROCESSOR_COUNT']
MAX_THREADS_PER_BLOCK = 128 # A manual choice

logger = logging.getLogger(__name__)

@dataclass(frozen=True)
class MMAAtom:
    name: str
    MNK: Tuple[int, ...]
    D_type: CudaDType
    A_type: CudaDType
    B_type: CudaDType
    C_type: CudaDType
    thread_count: int # Number of threads required to run the atom

MMAAtom_registry = {}

def register_mma_atom(name, MNK, A_type, B_type, C_type, D_type, thread_count):
    atom = MMAAtom(name, MNK, A_type, B_type, C_type, D_type, thread_count)
    MMAAtom_registry[name] = atom


@dataclass
class Config:
    MNKP: Tuple[int, ...]
    MNKTileShape: Shape | Tuple
    MNKAtomPlacement: Shape | Tuple
    Atom: MMAAtom
    dtype: th.dtype
    acc_dtype: th.dtype
    smempipe: int = 1
    regpipe: int = 1
    use_ldsm: bool = True
    swizzle: int = 0
    max_smem: int = SMEM_LIMIT
    max_threads: int = MAX_THREADS_PER_BLOCK

    @property
    def thread_count(self):
        return self.Atom.thread_count * prod(self.MNKAtomPlacement) # type: ignore

# The fundamental MMA atoms based on FMA that work on all architectures for all problem shapes
ALL_FLOAT_TYPES = [CudaDType.FLOAT32, CudaDType.FLOAT16, CudaDType.BFLOAT16]
for A_dtype in ALL_FLOAT_TYPES:
    for B_dtype in ALL_FLOAT_TYPES:
        for C_dtype in ALL_FLOAT_TYPES:
            for D_dtype in ALL_FLOAT_TYPES:
                register_mma_atom(f"UniversalFMA<{A_dtype.value},{B_dtype.value},{C_dtype.value},{D_dtype.value}>", (1, 1, 1), A_dtype, B_dtype, C_dtype, D_dtype, 1)

# A list of all available MMA atoms for the SM80 architecture
register_mma_atom("SM80_16x8x8_F16F16F16F16_TN", (16, 8, 8), CudaDType.FLOAT16, CudaDType.FLOAT16, CudaDType.FLOAT16, CudaDType.FLOAT16, 32)
register_mma_atom("SM80_16x8x16_F16F16F16F16_TN", (16, 8, 16), CudaDType.FLOAT16, CudaDType.FLOAT16, CudaDType.FLOAT16, CudaDType.FLOAT16, 32)
register_mma_atom("SM80_16x8x8_F32F16F16F32_TN", (16, 8, 8), CudaDType.FLOAT32, CudaDType.FLOAT16, CudaDType.FLOAT16, CudaDType.FLOAT32, 32)
register_mma_atom("SM80_16x8x16_F32F16F16F32_TN", (16, 8, 16), CudaDType.FLOAT32, CudaDType.FLOAT16, CudaDType.FLOAT16, CudaDType.FLOAT32, 32)
register_mma_atom("SM80_16x8x8_F32BF16BF16F32_TN", (16, 8, 8), CudaDType.FLOAT32, CudaDType.BFLOAT16, CudaDType.BFLOAT16, CudaDType.FLOAT32, 32)
register_mma_atom("SM80_16x8x16_F32BF16BF16F32_TN", (16, 8, 16), CudaDType.FLOAT32, CudaDType.BFLOAT16, CudaDType.BFLOAT16, CudaDType.FLOAT32, 32)
register_mma_atom("SM80_16x8x4_F32TF32TF32F32_TN", (16, 8, 4), CudaDType.FLOAT32, CudaDType.FLOAT32, CudaDType.FLOAT32, CudaDType.FLOAT32, 32)
register_mma_atom("SM80_16x8x8_F32TF32TF32F32_TN", (16, 8, 8), CudaDType.FLOAT32, CudaDType.FLOAT32, CudaDType.FLOAT32, CudaDType.FLOAT32, 32)
register_mma_atom("SM80_8x8x4_F64F64F64F64_TN", (8, 8, 4), CudaDType.FLOAT64, CudaDType.FLOAT64, CudaDType.FLOAT64, CudaDType.FLOAT64, 32)
# struct SM80_8x8x4_C64C64C64C64_TN
# struct SM80_8x8x4_GC64C64C64GC64_TN
# struct SM80_8x8x16_S32S8S8S32_TN
# struct SM80_8x8x16_S32S8S8S32_TN_SATURATE
# struct SM80_16x8x16_S32S8S8S32_TN
# struct SM80_16x8x16_S32S8S8S32_TN_SATURATE
# struct SM80_16x8x32_S32S8S8S32_TN
# struct SM80_16x8x32_S32S8S8S32_TN_SATURATE
# struct SM80_8x8x16_S32S8U8S32_TN
# struct SM80_8x8x16_S32S8U8S32_TN_SATURATE
# struct SM80_16x8x16_S32S8U8S32_TN
# struct SM80_16x8x16_S32S8U8S32_TN_SATURATE
# struct SM80_16x8x32_S32S8U8S32_TN
# struct SM80_16x8x32_S32S8U8S32_TN_SATURATE
# ......... there are more but we don't need them for now


def get_largest_atoms(atoms):
    """Returns the list of the atoms with the largest MNK.
    Where atom0.MNK < atom1.MNK iff atom0.MNK[0] < atom1.MNK[0] and atom0.MNK[1] < atom1.MNK[1] and atom0.MNK[2] < atom1.MNK[2]
    Since this relationship is a partial order we only filter out the atoms that are strictly smaller than others"""
    def is_smaller(atom0, atom1):
        # If any dimension of atom0 is greater, then atom0 is not strictly smaller
        if atom0.MNK[0] > atom1.MNK[0] and atom0.MNK[1] > atom1.MNK[1] and atom0.MNK[2] > atom1.MNK[2]:
            return False
        # If at least one dimension of atom0 is smaller, then atom0 is strictly smaller
        return atom0.MNK[0] < atom1.MNK[0] or atom0.MNK[1] < atom1.MNK[1] or atom0.MNK[2] < atom1.MNK[2]
    largest_atoms = set()
    for atom in atoms:
        # If atoms is smaller than any of the largest atoms, skip it
        if any(is_smaller(atom, largest_atom) for largest_atom in largest_atoms):
            continue
        # remove any ellements in the largest atoms set that are strictly smaller than atom
        for largest_atom in list(largest_atoms):
            if is_smaller(largest_atom, atom):
                largest_atoms.remove(largest_atom)
        largest_atoms.add(atom)
    return list(largest_atoms)

def evenly_divides(a, b):
    return all([a % b == 0 for a, b in zip(a, b)])

def get_valid_atoms(MNKTileShape, dtype, acc_dtype):
    """Returns all the atoms that evenly divide the tile shape and which have compatible types
    """
    Dtype = torch_dtype_to_c(dtype)
    AccDtype = torch_dtype_to_c(acc_dtype)
    valid_atoms = []
    for atom in MMAAtom_registry.values():
        divides = evenly_divides(MNKTileShape, atom.MNK)
        valid_types = atom.A_type.value == Dtype and atom.B_type.value == Dtype and atom.C_type.value == AccDtype and atom.D_type.value == AccDtype
        if divides and valid_types:
            valid_atoms.append(atom)
    return valid_atoms

def reasonable_MNK_tile_shapes(M=None, N=None, K=None, MNK=None):
    """Generate a list of reasonable tile shapes for the MMA problem.
    """
    default_sizes = {1, 2, 4, 8, 16, 32, 64, 128}
    K_default_sizes = {1, 2, 4, 8, 16, 32, 64}
    M_sizes = default_sizes if M is None else {M,}
    N_sizes = default_sizes if N is None else {N,}
    K_sizes = K_default_sizes if K is None else {K,}
    for M in list(M_sizes):
        for N in list(N_sizes):
            for K in list(K_sizes):
                yield (M, N, K)

def reasonable_atom_placements(MNKTileShape, atom, max_threads):
    """Generate a list of reasonable atom placements for the MMA problem.
    """
    M_rest = MNKTileShape[0] // atom.MNK[0]
    N_rest = MNKTileShape[1] // atom.MNK[1]
    # Only loop over powers of 2
    for M in [2**i for i in range(0, M_rest.bit_length()) if 2**i <= M_rest]:
        for N in [2**j for j in range(0, N_rest.bit_length()) if 2**j <= N_rest]:
            if M * N * atom.thread_count > max_threads:
                continue
            yield (M, N, 1)

def dtype_to_bytes(dtype):
    if dtype in [th.float64, th.int64]:
        return 8
    elif dtype in [th.float32, th.int32]:
        return 4
    elif dtype in [th.float16, th.bfloat16, th.int16, th.uint16]:
        return 2
    elif dtype in [th.int8, th.uint8]:
        return 1
    raise ValueError(f"Unsupported dtype: {dtype}")

def safe_filter(fn):
    @wraps(fn)
    def wrapper(*args, **kwargs):
        arg_names = list(inspect.signature(fn).parameters.keys())
        n_args = dict(zip(arg_names, args))
        old_configs = copy(n_args['configs'])
        new_configs = fn(*args, **kwargs)
        if not new_configs:
            logger.warning(f"No configs passed {fn.__name__} filter, returning all {len(old_configs)} configs")
            return old_configs
        return new_configs
    return wrapper

def mma_smem_estimate(MNKTileShape, dtype, acc_dtype, smempipe=1):
    A_smem = MNKTileShape[0] * MNKTileShape[2] * dtype_to_bytes(dtype) * smempipe
    B_smem = MNKTileShape[1] * MNKTileShape[2] * dtype_to_bytes(dtype) * smempipe
    C_smem = MNKTileShape[0] * MNKTileShape[1] * dtype_to_bytes(acc_dtype)
    return max(A_smem + B_smem, C_smem)

def initial_config_sweep(MNKP, dtype, acc_dtype, M_tile=None, N_tile=None, K_tile=None, max_smem=SMEM_LIMIT, max_threads=MAX_THREADS_PER_BLOCK) -> list[Config]:
    configs = []
    MNK = MNKP[:-1]
    # Generate an initial large list of reasonable configs
    for MNKTileShape in reasonable_MNK_tile_shapes(M_tile, N_tile, K_tile, MNK):
        if not evenly_divides(MNK, MNKTileShape):
            continue
        # Don't consider configs that use too much shared memory
        smem = mma_smem_estimate(MNKTileShape, dtype, acc_dtype)
        if smem > max_smem:
            continue
        for atom in get_valid_atoms(MNKTileShape, dtype, acc_dtype):
            for MNKAtomPlacement in reasonable_atom_placements(MNKTileShape, atom, max_threads):
                configs.append(Config(
                    MNKTileShape=MNKTileShape,
                    MNKAtomPlacement=MNKAtomPlacement,
                    Atom=atom,
                    dtype=dtype,
                    acc_dtype=acc_dtype,
                    MNKP=MNKP,
                    max_smem=max_smem,
                    max_threads=max_threads,
                ))
    logger.info(f"Generated {len(configs)} initial mma configs")
    return configs

@safe_filter
def thread_count_filter(configs):
    # If the candiate configs don't even use a full warp, pick the largerst one (relevant for small problems)
    thread_counts = [c.thread_count for c in configs]
    max_thread_count = max(thread_counts)
    if max_thread_count <= 32:
        configs = [c for c, t in zip(configs, thread_counts) if t == max_thread_count]

    # If there are configs that use multiples of 32 threads filter out the rest
    if any([t % 32 == 0 for t in thread_counts]):
        configs = [c for c, t in zip(configs, thread_counts) if t % 32 == 0]
    logger.info(f"Filtered to {len(configs)} mma configs due to thread count")
    return configs

@safe_filter
def atom_size_filter(configs):
    # Use the largest possible atom. Avoid benchmarking smaller instructions. TODO: verify this is a good idea
    largest_atoms = get_largest_atoms([c.Atom for c in configs])
    configs = [c for c in configs if c.Atom in largest_atoms]
    logger.info(f"Filtered to {len(configs)} mma configs for largest atoms")
    return configs

@safe_filter
def saturation_filter(configs):
    MNKP = configs[0].MNKP
    # If the problem does not saturate the GPU, pick the config with the largest number of C tiles
    batch_size = MNKP[3] if isinstance(MNKP[3], int) else prod(MNKP[3])
    C_tile_nums = [ (MNKP[0]//c.MNKTileShape[0]) * (MNKP[1]//c.MNKTileShape[1]) * batch_size for c in configs]
    max_C_tile_num = max(C_tile_nums)
    if max_C_tile_num < SM_NUMBER:
        configs = [c for c, C_tile_num in zip(configs, C_tile_nums) if C_tile_num == max_C_tile_num]
    logger.info(f"Filtered to {len(configs)} mma configs for C tile count")
    return configs

@safe_filter
def arithmetic_intensity_filter(min_MN_tile_size, configs):
    # We want to pick as large a MNTile as possible to imporove arithmetic intensity
    MN_tile_sizes = [c.MNKTileShape[0] * c.MNKTileShape[1] for c in configs]
    if any([MN_tile_size > min_MN_tile_size for MN_tile_size in MN_tile_sizes]):
        configs = [c for c, MN_tile_size in zip(configs, MN_tile_sizes) if MN_tile_size >= min_MN_tile_size]
    logger.info(f"Filtered to {len(configs)} mma configs for large MN tile size")
    return configs

def performance_expansion(configs: List[Config], smempipe=None, regpipe=None, use_ldsm=None, swizzle=None):
    smempipes = [1, 2] if smempipe is None else (smempipe,) if not isinstance(smempipe, Iterable) else smempipe
    regpipes = [1] if regpipe is None else (regpipe,) if not isinstance(regpipe, Iterable) else regpipe
    use_ldsms = [True] if use_ldsm is None else (use_ldsm,) if not isinstance(use_ldsm, Iterable) else use_ldsm
    swizzles = [1] if swizzle is None else (swizzle,) if not isinstance(swizzle, Iterable) else swizzle
    new_configs = []
    for config in configs:
        for smempipe, regpipe, use_ldsm, swizzle in product(smempipes, regpipes, use_ldsms, swizzles):
            c = copy(config)
            c.smempipe = smempipe
            c.regpipe = regpipe
            c.use_ldsm = use_ldsm
            c.swizzle = swizzle
            if mma_smem_estimate(c.MNKTileShape, c.dtype, c.acc_dtype, smempipe) > SMEM_LIMIT:
                continue
            new_configs.append(c)
    logger.info(f"Expanded to {len(new_configs)} configs with smempipe, regpipe, use_ldsm, swizzle for mma")
    return new_configs

def render_configs(configs: List[Config]):
    # ---- Massage the data so it's compatible with the kernel ----
    # We dont want to return full atom object, just the name
    new_configs = [
        {
            'Atom': c.Atom.name,
            'MNKTileShape': static_shape(c.MNKTileShape),
            'MNKAtomPlacement': static_shape(c.MNKAtomPlacement),
            **{k: v for k, v in c.__dict__.items() if k in ['smempipe', 'regpipe', 'use_ldsm', 'swizzle']}
        }
        for c in configs
    ]
    logger.info(f"Filtered to {len(new_configs)} mma configs for kernel compatibility")
    return new_configs

def register_estimate(config: Config):
    # A rough and conservative estimate of the number of registers needed
    tile = config.MNKTileShape
    a_reg = tile[0] * tile[2] # type: ignore
    b_reg = tile[1] * tile[2] # type: ignore
    c_reg = tile[0] * tile[1] # type: ignore
    return float(a_reg + b_reg + c_reg) # type: ignore

def register_per_thread_estimate(config: Config):
    # A rough and conservative estimate of the number of registers needed per thread
    tile = config.MNKTileShape
    atom = config.Atom
    placement = config.MNKAtomPlacement
    rest_m, rest_n, rest_k = tile[0] // (atom.MNK[0] * placement[0]), tile[1] // (atom.MNK[1] * placement[1]), tile[2] // (atom.MNK[2] * placement[2]) # type: ignore
    a_reg = atom.MNK[0] * atom.MNK[2] * rest_m * rest_k // 32
    b_reg = atom.MNK[1] * atom.MNK[2] * rest_n * rest_k // 32
    c_reg = atom.MNK[0] * atom.MNK[1] * rest_m * rest_n // 32

    return float(a_reg + b_reg + c_reg) # type: ignore

@safe_filter
def register_spillage_filter(configs):
    # We want to avoid register spillage both on per thread level and on per SM level
    configs = [c for c in configs if register_estimate(c) <= REG_PER_SM and register_per_thread_estimate(c) <= 256]
    logger.info(f"Filtered to {len(configs)} mma configs for register spillage")
    return configs

@safe_filter
def register_underutil_filter(level, configs):
    # We want to avoid underutilizing the registers, if we are using < x% of
    # max registers, we filter out the configs that are underutilized
    top_usage = sorted([register_estimate(c) for c in configs], reverse=True)[0]
    configs = [c for c in configs if register_estimate(c) >= top_usage * level]
    logger.info(f"Filtered to {len(configs)} mma configs for register underutilization")
    return configs

@safe_filter
def regpipe_filter(regpipes, configs):
    # Filter out configs that use regpipe other than the ones we want to use
    configs = [c for c in configs if c.regpipe in regpipes]
    logger.info(f"Filtered to {len(configs)} mma configs for regpipe")
    return configs

@safe_filter
def smempipe_filter(smempipes, configs):
    # Filter out configs that use smempipe other than the ones we want to use
    configs = [c for c in configs if c.smempipe in smempipes]
    logger.info(f"Filtered to {len(configs)} mma configs for smempipe")
    return configs

@safe_filter
def warp_filter(num_warps, configs):
    # Filter out configs that use more warps than the number of warps we want to use
    configs = [c for c in configs if prod(c.MNKAtomPlacement) <= num_warps]
    logger.info(f"Filtered to {len(configs)} mma configs for warp size")
    return configs

def make_configurator(filter_funcs):

    @lru_cache(maxsize=None)
    def mma_config_sweep(MNKP, dtype, acc_dtype, M_tile=None, N_tile=None, K_tile=None, smempipe=None, regpipe=None, use_ldsm=None, swizzle=None, max_smem=SMEM_LIMIT, max_threads=MAX_THREADS_PER_BLOCK) -> List[Dict[str, Any]]:
        """Given an MMA problem, return a list of configs that can be used to configure any of the MMA kernels.
        Args:
            MNK: The shape of the MMA problem.
            dtype: The data type of the A, B matrices.
            acc_dtype: The data type of the C accumulator. (could be different from the C.dtype)
            M_tile (optional): The M_tile size
            N_tile (optional): The N_tile size
            K_tile (optional): The K_tile size
        Each returned config is a dict with keys:
            - MNKTileShape: The shape of the tile to use for the MMA kernel. Must divide the problem shape
            - Atom: The atom (MMA instruction) to use
            - MNKAtomPlacement: How to replicate the atoms across more threads of the CTA.
            Eg, if the instruction is warp level, then MNKAtomPlacement will control the number of warps in the CTA.
        
        We try to filter out as many valid configs as possible to speed up the sweeps. For example
            - We only return configs that use the largest possible atoms. If a 16x8x16 atom is valid we don't need to consider 16x8x8 or FMA based matmuls
            - We can filter out any configs that use too much shared memory
            - For small problems using < 32 threads, we return configs that use the maximum number of threads
            - Since the size of the MN tile controls the amount of gmem reloading, we only consider MN tiles >= 1024 if the setting allows it
        """
        configs = initial_config_sweep(MNKP, dtype, acc_dtype, M_tile, N_tile, K_tile, max_smem, max_threads)
        configs = performance_expansion(configs, smempipe, regpipe, use_ldsm, swizzle)
        for filter_func in filter_funcs:
            configs = filter_func(configs)

        configs = render_configs(configs)
        return configs
    
    return mma_config_sweep
    

# The following configurators are heuristics used to prune configs that are likely to have bad performance
# However there's no guarantee that the the top-performing config is not filtered out.
no_filter_configurator = make_configurator([])

BASE_FILTERS = [
    thread_count_filter, # For small problems, we want to use the maximum number of threads
    atom_size_filter, # Use the largest possible atom. Avoid benchmarking smaller instructions.
    register_spillage_filter, # Avoid register spillage
    saturation_filter, # If the problem does not saturate the GPU, pick the config with the largest number of C tiles
    partial(arithmetic_intensity_filter, 2048), # We want to pick as large a MNTile as possible to imporove arithmetic intensity
]

base_configurator = make_configurator(BASE_FILTERS)

ADVANCED_FILTERS = [
    thread_count_filter,
    atom_size_filter,
    register_spillage_filter, # Avoid register spillage
    saturation_filter,
    partial(arithmetic_intensity_filter, 2048),
    partial(register_underutil_filter, 0.4), # Avoid underutilizing the registers relative to maximum register usage
    partial(warp_filter, 4), # Avoid using more warps than the number of warps we want to use
    partial(smempipe_filter, [1, 2]), # Smempipe 1 and 2 are the only ones we want to use
    partial(regpipe_filter, [1]), # Regpipe 1 is probably good enough
]

advanced_configurator = make_configurator(ADVANCED_FILTERS)







from math import sqrt
import pytest
import torch as th
from vidrial.jit.static.shape import Shape
from vidrial.py_utils.test_utils import diff
from vidrial.kernels.mma.binding import binding
from vidrial.kernels.mma.op import op, op_reference


@pytest.mark.parametrize("P,M,N,K", [
    (1, 16, 8, 16),
    (4, 16, 8, 16),
])
@pytest.mark.parametrize("dtype", [
    th.float16,
    th.bfloat16,
    th.float32, # kernel error with float32 & tf32
])
def test_binding_pickbest(P, M, N, K, dtype):
    """Test raw_mma kernel on GPU."""
    A = th.randn(P, M, K, dtype=dtype, device="cuda") / sqrt((M+K)/2)
    B = th.randn(P, K, N, dtype=dtype, device="cuda") / sqrt((K+N)/2)
    C = th.empty(P, M, N, dtype=dtype, device="cuda") / sqrt((M+N)/2)
    binding(A, B, C)
    C_ref = A @ B
    diff(C, C_ref, rtol=1e-1, atol=1e-1)

@pytest.mark.parametrize("P,M,N,K", [
    (1, 16, 64, 32),
    (16, 32, 32, 32),
])
@pytest.mark.parametrize("dtype", [
    th.float16,
    th.bfloat16,
    th.float32, # kernel error with float32 & tf32
])
def test_binding(P, M, N, K, dtype):
    """Test raw_mma kernel with sweep of tileshapes and atom placements"""
    if dtype == th.float16:
        atom = "SM80_16x8x8_F32F16F16F32_TN"
    elif dtype == th.bfloat16:
        atom = "SM80_16x8x8_F32BF16BF16F32_TN"
    elif dtype == th.float32:
        atom = "SM80_16x8x4_F32TF32TF32F32_TN"
    else:
        raise ValueError(f"Unsupported dtype: {dtype}")
    configs = [
        {
            "MNKTileShape": Shape(16, 32, 16),
            "MNKAtomPlacement": Shape(1, 1, 1),
            "Atom": atom,
        },
        {
            "MNKTileShape": Shape(16, 16, 16),
            "MNKAtomPlacement": Shape(1, 1, 1),
            "Atom": atom,
        },
    ]
    A = th.randn(P, M, K, dtype=dtype, device="cuda") / sqrt((M+K)/2)
    B = th.randn(P, K, N, dtype=dtype, device="cuda") / sqrt((K+N)/2)
    C = th.empty(P, M, N, dtype=dtype, device="cuda") / sqrt((M+N)/2)
    C_ref = A @ B
    for config in configs:
        C[:] = 0
        binding(A, B, C, config["Atom"], config["MNKTileShape"], config["MNKAtomPlacement"], 1, 1, True, 1)
        diff(C, C_ref, rtol=1e-1, atol=1e-1)


@pytest.mark.parametrize("batch_dims,M,N,K", [
    ((), 16, 16, 16),
    ((4,), 64, 64, 64),
    ((8,), 32, 64, 32),
    ((16,), 64, 32, 64),
])
def test_mma_autograd_against_reference(batch_dims, M, K, N):
    """Test that mma gradients match PyTorch's native implementation."""
    dtype = th.float16
    device = 'cuda'
    generator = th.Generator(device=device).manual_seed(42)
    A = th.randn(batch_dims + (M, K), dtype=dtype, device=device, generator=generator, requires_grad=True)
    B = th.randn(batch_dims + (K, N), dtype=dtype, device=device, generator=generator, requires_grad=True)
    dC = th.randn(batch_dims + (M, N), dtype=dtype, device=device, generator=generator, requires_grad=True)
    # Custom op fwd and bwd 
    C = op(A, B)
    C.backward(dC)
    assert A.grad is not None and B.grad is not None
    dA = A.grad.clone()
    dB = B.grad.clone()
    A.grad, B.grad = None, None
    # Reference op fwd and bwd 
    C_ref = op_reference(A, B)
    C_ref.backward(dC)
    assert A.grad is not None and B.grad is not None
    dA_ref = A.grad.clone()
    dB_ref = B.grad.clone()
    # Compare results
    diff(C, C_ref, rtol=1e-1, atol=1e-1)
    diff(dA, dA_ref, rtol=1e-1, atol=1e-1)
    diff(dB, dB_ref, rtol=1e-1, atol=1e-1)



import torch as th
from vidrial.kernels.flash.binding import binding

#TODO: support GQA, non-causal, etc
@th.library.custom_op("vidrial::flash", mutates_args=())
def op(Q: th.Tensor, K: th.Tensor, V: th.Tensor, softmax_scale: float = None) -> tuple[th.Tensor, th.Tensor]:
    """ Computes flash attention

    Arguments:
        Q: (batch_size, seqlen_q, nheads, head_dim), the query tensor
        K: (batch_size, seqlen_k, nheads, head_dim), the key tensor    
        V: (batch_size, seqlen_k, nheads, value_dim), the value tensor
        softmax_scale: float. The scaling of QK^T before applying softmax.
            Default to 1 / sqrt(head_dim).

    Returns:
        O: (batch_size, seqlen_q, nheads, value_dim), the output tensor
        l: (batch_size, seqlen_q, nheads), the log-sum-of-exponentials factors
    """
    batch_size, seqlen, nheads, head_dim, value_dim = *Q.shape, V.shape[-1]
    assert K.shape == (batch_size, seqlen, nheads, head_dim), f"Invalid K shape: {K.shape}"
    assert V.shape == (batch_size, seqlen, nheads, value_dim), f"Invalid V shape: {V.shape}"

    if softmax_scale is None:
        softmax_scale = 1.0 / head_dim**0.5
    O = th.empty(Q.shape[:-1] + (value_dim,), device=Q.device, dtype=Q.dtype)
    l = th.empty(Q.shape[:-1], device=Q.device, dtype=th.float32)
    Q, K, V, O, l = map(lambda x: x.transpose(1, 2), (Q, K, V, O, l)) # [b, h, t, d]
    binding(Q, K, V, O, l, softmax_scale) # type: ignore
    O, l = map(lambda x: x.transpose(1, 2), (O, l)) # [b, t, h, d]
    return O, l

@op.register_fake
def op_fake(Q: th.Tensor, K: th.Tensor, V: th.Tensor, softmax_scale: float = None):
    return th.empty(Q.shape[:-1] + (V.shape[-1],), device=Q.device, dtype=Q.dtype), th.empty(Q.shape[:-1], device=Q.device, dtype=th.float32)


def op_reference(Q, K, V, O, l):
    S = (Q @ K.transpose(-1, -2)).to(th.float32)
    b, tq, tk = S.shape
    M = th.tril(th.ones(tq, tk, device=S.device, dtype=th.bool)).unsqueeze(0).expand(b, -1, -1)
    S = S.masked_fill(~M, float('-inf'))
    S = S - S.max(dim=-1, keepdim=True).values
    P = th.exp(S)
    l = P.sum(dim=-1)
    P = (P / l.unsqueeze(-1)).to(O.dtype)
    O = P @ V
    return O, l


import torch as th
from vidrial.kernels.mma.binding import binding
from vidrial.kernels.mma.dimensions import problem_shape

def canonical_inputs(A, B):
    """ Make all the arguments 3D and return a function to convert the output to the original shape """
    assert A.device == B.device, "A and B must be on the same device"
    assert A.dtype == B.dtype, "A and B must have the same dtype"
    assert A.ndim == B.ndim
    assert A.ndim in [2, 3]
    expand_batch = A.ndim == 2
    if expand_batch: A, B = A.unsqueeze(0), B.unsqueeze(0)
    problem_shape(A.shape, B.shape) # checks that the dimensions are valid
    return A, B, lambda C: C[0] if expand_batch else C

@th.library.custom_op("vidrial::mma", mutates_args=())
def op(A: th.Tensor, B: th.Tensor) -> th.Tensor:
    r"""Compute matrix multiplication: A @ B.

    Args:
        A: Input tensor of shape `(batch_size, m, k)`.
           The left matrix in the multiplication.
        B: Input tensor of shape `(batch_size, k, n)`.
           The right matrix in the multiplication.

    Returns:
        C: Output tensor of shape `(batch_size, m, n)`.
           The result of A @ B.

    Note:
        - All input tensors must be on the same device
        - All input tensors must have the same dtype
        - The batch dimension is treated as a batch dimension
        - The operation is optimized for CUDA devices
    """
    A, B, cannonical_outputs = canonical_inputs(A, B)
    P, M, N, K = problem_shape(A.shape, B.shape)
    C = th.empty(P, M, N, device=A.device, dtype=A.dtype)
    binding(A, B, C)
    return cannonical_outputs(C)

# Make it traceable
@op.register_fake
def op_fake(A: th.Tensor, B: th.Tensor):
    A, B, cannonical_outputs = canonical_inputs(A, B)
    P, M, N, K = problem_shape(A.shape, B.shape)
    C = th.empty(P, M, N, device=A.device, dtype=A.dtype)
    return cannonical_outputs(C)


# ---------------------- Autograd --------------------------------

def op_setup(ctx, inputs, output):
    A, B = inputs
    ctx.save_for_backward(A, B)
def op_backward(ctx, dC):
    A, B = ctx.saved_tensors
    dA = op(dC, B.transpose(-1, -2))
    dB = op(A.transpose(-1, -2), dC)
    return dA, dB
th.library.register_autograd(
    "vidrial::mma", op_backward, setup_context=op_setup
)


# ---------------------- Reference --------------------------------

def op_reference(A: th.Tensor, B: th.Tensor) -> th.Tensor:
    return A @ B



from math import comb, factorial
import torch as th

def problem_dimensions(Z_shape, X_shape, power, d_tile=1):
    assert Z_shape == op_output_shape(X_shape, power, d_tile)
    bs, d = X_shape[:-1], X_shape[-1]
    assert d % d_tile == 0, f"d={d} must be divisible by d_tile={d_tile}"
    D = sympow_dim(d, power, d_tile)
    return bs, d, D

def sympow_dim(d, power, d_tile=1):
    if d_tile == 1:
        return comb(d + power - 1, power)
    return sympow_dim(d // d_tile, power) * d_tile**power

def op_output_shape(x_shape, power, d_tile=1):
    batch_dims, d = tuple(x_shape[:-1]), x_shape[-1]
    assert d % d_tile == 0, f"d={d} must be divisible by d_tile={d_tile}"
    d_tile_num = sympow_dim(d // d_tile, power)
    return th.Size(batch_dims + (d_tile_num,) + (d_tile,) * power)
   
def interface_output_shape(X_shape, power, d_tile=1, dim=-1):
    Z_shape = list(X_shape)
    Z_shape[dim] = sympow_dim(X_shape[dim], power, d_tile)
    return th.Size(Z_shape)

sympow_shape = interface_output_shape
sympow_op_shape = op_output_shape

def canonical_arguments(X):
    if len(X.shape) == 1: X = X.unsqueeze(0)
    assert len(X.shape) == 2, f"X_shape={X.shape} must be 2D. Generic batch dims not implemented yet"
    return X

class SympowCoords:
    def __init__(self, n, power):
        assert n >= 1 and power >= 1, f"n={n} and power={power} must be >= 1"
        self.n, self.power = n, power
        self.seq, self.idx = [0] * power, 0
        self.final = False
        
    def duplicate_count(self):
        hist = [0]*self.n
        for i in self.seq:
            hist[i] += 1
        count = factorial(self.power)
        for h in hist:
            count //= factorial(h)
        return count
        
    def increment(self, position):
        assert position >= 0 and position < self.power, f"position={position} must be >= 0 and < power={self.power}"
        if self.seq[position] < self.n-1:
            self.idx += 1
            self.seq[:position+1] = [self.seq[position]+1] * (position+1)
            return
        if position == self.power-1:
            self.final = True
            return
        self.increment(position+1)
        
    def __iter__(self): return self
    
    def __next__(self):
        if self.final: raise StopIteration
        result = (self.idx, self.seq.copy(), self.duplicate_count())
        self.increment(0)
        return result
        
    def __len__(self):
        return sympow_dim(self.n, self.power)




from typing import Dict, Any
import torch as th
from dataclasses import dataclass
from functools import partial
import logging
from vidrial.kernels.sympow.dimensions import sympow_dim
from vidrial.jit.static import Shape, Layout, Int, Tuple
from vidrial.jit.static.util import layout_from_shape_stride, torch_dtype_to_c
from vidrial.kernels.mma_configurator import make_configurator as create_configurator, dtype_to_bytes, ADVANCED_FILTERS, safe_filter, Config, CudaDType
from vidrial.jit.jit import jit, render
from vidrial.jit.timingcache import ConfigTimingCache
from vidrial.jit.binding import make_binding
from vidrial.kernels.mma_sympow_bwd.dimensions import canonical_inputs, dimensions

# ------------------- Source Code -------------------
@dataclass
class SourceCode:
    T: str
    MNKPSlabShape: Shape
    d: int
    d_tile: int
    pow: int
    GASlab: Layout
    GBSlab: Layout
    GcSlab: Layout
    duplicate_correction: bool
    MNKTileShape: Shape
    MNKAtomPlacement: Shape
    Atom: str
    smempipe: int
    regpipe: int
    use_ldsm: bool
    swizzle: bool
    SmemAccI: Tuple
    scale_A: bool
    @property
    def template(self):
        return """
#include <cuda.h>
#include <cuda_runtime.h>
#include <iostream>
#include "cute/tensor.hpp"
#include "kernels/mma_sympow_bwd/kernel.cuh"
using namespace cute;
using namespace vidrial;
extern "C" int launch(void* __raw_A, void* __raw_B, void* __raw_c, void* __raw_c_dot, float sA) {
    using T = {T};
    using MNKPSlabShape = decltype(static_tree_cast<int64_t>({MNKPSlabShape}{}));
    using MNKTileShape = decltype(static_tree_cast<int64_t>({MNKTileShape}{}));
    constexpr int d = {d}, d_tile = {d_tile};
    constexpr int pow = {pow};
    constexpr int expand_dim = -2;
    using GASlab = decltype(static_tree_cast<int64_t>({GASlab}{}));
    using GBSlab = decltype(static_tree_cast<int64_t>({GBSlab}{}));
    using GcSlab = decltype(static_tree_cast<int64_t>({GcSlab}{}));
    constexpr bool duplicate_correction = {duplicate_correction};
    using PCfg = PerfCfg<{smempipe}, {regpipe}, {use_ldsm}, {swizzle}>;
    using Atom = MMA_Atom<{Atom}>;
    using MNKAtomPlacement = decltype(static_tree_cast<int64_t>({MNKAtomPlacement}{}));
    using SmemAccI = {SmemAccI};
    using KernelCfg = SympowCMmaCfg<T, pow, Atom, MNKAtomPlacement, MNKPSlabShape, MNKTileShape, d, d_tile, GASlab, GBSlab, GcSlab, SmemAccI, PCfg>;
    auto A = reinterpret_cast<T*>(__raw_A);
    auto B = reinterpret_cast<T*>(__raw_B);
    auto c = reinterpret_cast<T*>(__raw_c);
    auto c_dot = reinterpret_cast<T*>(__raw_c_dot);
    return launch_mma_sympow_bwd_kernel<duplicate_correction, {scale_A}>(KernelCfg{}, A, B, c, c_dot, sA);
}
"""
    def __str__(self):
        return render(self.template, self.__dict__)

# ------------------- Binding -------------------
@dataclass
class BindingCfg:
    """ The main methods:
          from_args: constructs a BindingCfg with the binding arguments themselves
          source: the source code object that the binding relies upon
          canonical_args: preprocesses the arguments into the form accepted by the source code
    """
    batches: tuple[int, ...]
    M: int
    N: int
    K: int
    d: int
    D: int
    dtype: th.dtype
    A_shape: tuple[int, ...]
    B_shape: tuple[int, ...]
    c_shape: tuple[int, ...]
    A_stride: tuple[int, ...]
    B_stride: tuple[int, ...]
    c_stride: tuple[int, ...]
    expand_dim: int  # TODO: Currently only expand_dim=-1 (D=N) is supported
    power: int
    d_tile: int
    scale_A: bool
    duplicate_correction: bool
    MNKTileShape: Shape
    MNKAtomPlacement: Shape
    Atom: str
    smempipe: int
    regpipe: int
    use_ldsm: bool
    swizzle: bool
    SmemAccI: Tuple
    @classmethod
    def from_args(cls, A, B, c, c_dot, expand_dim, power, d_tile, scale_A, duplicate_correction, MNKTileShape, MNKAtomPlacement, Atom, smempipe, regpipe, use_ldsm, swizzle, SmemAccI):
        assert A.device.type == B.device.type == c.device.type == c_dot.device.type == 'cuda', f"Invalid {A.device}, {B.device}, {c.device}, {c_dot.device}."
        assert A.dtype == B.dtype == c.dtype
        assert c.shape == c_dot.shape and c.stride() == c_dot.stride() and c.dtype == c_dot.dtype, "c and c_dot should have same shape, stride and dtype"
        assert expand_dim == -2, "Currently only expand_dim=-2 is supported"
        dtype = A.dtype
        A, B, c, c_dot, expand_dim = canonical_inputs(A, B, c, c_dot, expand_dim)
        batches, M, K, N, d = dimensions(A.shape, B.shape, c.shape, power, d_tile)
        D = sympow_dim(d, power, d_tile)
        assert M == D, f"N={N} must be equal to sympow_dim(d, power, d_tile)={D}"
        A_shape, B_shape, c_shape = tuple(A.shape), tuple(B.shape), tuple(c.shape)
        A_stride, B_stride, c_stride = A.stride(), B.stride(), c.stride()
        return cls(batches, M, N, K, d, D, dtype,
                   A_shape, B_shape, c_shape, A_stride, B_stride, c_stride,
                   expand_dim, power, d_tile, scale_A != 1.0, duplicate_correction,
                   MNKTileShape, MNKAtomPlacement, Atom,
                   smempipe, regpipe, use_ldsm, swizzle, SmemAccI)
    @property
    def source(self):
        """ Conversion between the different variable name and shape conventions
                M = D = sympow_dim(d, power, d_tile)
                    Python                     c++
                A_shape=[P,M,K]      ->      aSlabShape=[M,K,P]
                B_shape=[P,K,N]      ->      BSlabShape=[N,K,P]
                c_shape=[P,d,N]      ->      CSlabShape=[d,N,P]
        """
        batch_inds = tuple(range(len(self.batches)))
        PShape = Shape(*[Int(b) for b in self.batches])
        return SourceCode(
            T=torch_dtype_to_c(self.dtype),
            MNKPSlabShape=Shape(Int(self.M), Int(self.N), Int(self.K), PShape),
            d=self.d,
            d_tile=self.d_tile,
            pow=self.power,
            GASlab=layout_from_shape_stride(self.A_shape, self.A_stride, (-2, -1, batch_inds)),
            GBSlab=layout_from_shape_stride(self.B_shape, self.B_stride, (-1, -2, batch_inds)),
            GcSlab=layout_from_shape_stride(self.c_shape, self.c_stride, (-2, -1, batch_inds)),
            duplicate_correction=self.duplicate_correction,
            MNKTileShape=self.MNKTileShape,
            MNKAtomPlacement=self.MNKAtomPlacement,
            Atom=self.Atom,
            smempipe=self.smempipe,
            regpipe=self.regpipe,
            use_ldsm=self.use_ldsm,
            swizzle=self.swizzle,
            SmemAccI=self.SmemAccI,
            scale_A=self.scale_A,
        )

# ------------------- Autotuned -------------------
def smem_estimate(d, power, d_tile, mma_config: Config):
    smempipe = mma_config.smempipe
    M_tile, N_tile, K_tile = mma_config.MNKTileShape
    dtype_bytes = dtype_to_bytes(mma_config.dtype)
    X_smem = N_tile * d * dtype_bytes
    X_grad_smem = X_smem
    A_smem, B_smem = M_tile * K_tile * smempipe * dtype_bytes, K_tile * N_tile * smempipe * dtype_bytes
    return X_smem + X_grad_smem + A_smem + B_smem

@safe_filter
def smem_filter(d, power, d_tile, configs):
    return [c for c in configs if smem_estimate(d, power, d_tile, c) <= c.max_smem]

def make_configurator(smempipe: tuple[int, int], regpipe: tuple[int, int], use_ldsm: bool, swizzle: int):
    """ returns function that constCreates a set of promising configurations for the MMA kernel
        The functional arguments are A,B,c,c_dot,expand_dim,power,d_tile,duplicate_correction
        The rest are performance arguments MNKTileShape, MNKAtomPlacement, Atom, smempipe, regpipe, use_ldsm, swizzle  """
    def configurator(args: dict) -> list[Dict[str, Any]]:
        A, B, c, c_dot, expand_dim = canonical_inputs(args['A'], args['B'], args['c'], args['c_dot'], args['expand_dim'])
        batches, M, K, N, d = dimensions(A.shape, B.shape, c.shape, args['power'], args['d_tile'])
        power, d_tile = args['power'], args['d_tile']
        D_tile = d_tile ** power
        configs = []
        mma_sympow_bwd_configurator = create_configurator(ADVANCED_FILTERS + [partial(smem_filter, d, power, d_tile)])
        cfgs = mma_sympow_bwd_configurator((M, N, K, batches), A.dtype, th.float32, D_tile, None, None, smempipe=smempipe, regpipe=regpipe, use_ldsm=use_ldsm, swizzle=swizzle)
        configs += cfgs
        SmemAccI = Tuple(*[Int(1) for i in range(power)])
        for cfg in configs:
            cfg['SmemAccI'] = SmemAccI
        return configs
    return configurator

def hash_fn(args: dict) -> str:
    A, B, c, c_dot, expand_dim = canonical_inputs(args['A'], args['B'], args['c'], args['c_dot'], args['expand_dim'])
    M, N, K, d = A.shape[-2], B.shape[-2], B.shape[-1], c.shape[-2]
    key = f"{A.dtype}-{B.dtype}-{c.dtype}-M_mod128_{M % 128}-N_mod128_{N % 128}-K_mod128_{K % 128}-d{d}-{expand_dim}-{args['power']}-{args['d_tile']}"
    return key

@make_binding(cache=ConfigTimingCache('mma_sympow_bwd', hash_fn),
          sweep=make_configurator(smempipe=(1,2), regpipe=(0,1), use_ldsm=True, swizzle=1))
def binding(A, B, c, c_dot, expand_dim, power, d_tile, scale_A, duplicate_correction, MNKTileShape, MNKAtomPlacement, Atom, smempipe, regpipe, use_ldsm, swizzle, SmemAccI):
    binding_cfg = BindingCfg.from_args(A, B, c, c_dot, expand_dim, power, d_tile, scale_A, duplicate_correction, MNKTileShape, MNKAtomPlacement, Atom, smempipe, regpipe, use_ldsm, swizzle, SmemAccI)
    name = "mma_sympow_bwd"
    logging.debug(f"Calling {name} with binding_cfg={binding_cfg}")
    return_code = jit(name=name, code=str(binding_cfg.source))(A, B, c, c_dot, scale_A)
    if return_code != 0:
        raise RuntimeError(f"Error during kernel execution. Failed with error code: {return_code}")




import torch
import math

def wrap_with_timer(fn, n=10, warmup=3):
    """Takes a function and returns a function that calls it n times and returns the total time."""
    def timed_fn(*args, **kwargs):
        torch.cuda.synchronize()
        for _ in range(warmup):
            fn(*args, **kwargs)

        cache = torch.empty(int(256e6), dtype=torch.int8, device='cuda')
        def flush_cache():
            cache.zero_()
        torch.cuda.synchronize()

        start_events = [torch.cuda.Event(enable_timing=True) for _ in range(n)]
        end_events = [torch.cuda.Event(enable_timing=True) for _ in range(n)]
        for i in range(n):
            flush_cache()
            torch.cuda._sleep(1_000_000_0)
            start_events[i].record() # type: ignore
            out = fn(*args, **kwargs)
            end_events[i].record() # type: ignore
        torch.cuda.synchronize()
        times = [s.elapsed_time(e) for s, e in zip(start_events, end_events)]
        total_time = sum(times)
        std = math.sqrt(sum((t - total_time / n) ** 2 for t in times) / n)
        return out, total_time, std
    return timed_fn


def timeit(fn, *args, num1=10, num2=50, **kwargs):
    """Takes a function and returns a an estimate of time per iteration."""
    timed_fn_1 = wrap_with_timer(fn, num1)
    timed_fn_2 = wrap_with_timer(fn, num2)

    _, t1, std1 = timed_fn_1(*args, **kwargs)
    _, t2, std2 = timed_fn_2(*args, **kwargs)

    return (t2 - t1) / (num2 - num1), math.sqrt((num1 * std1**2 + num2 * std2**2) / (num1 + num2))


import torch
import logging
import pandas as pd
from functools import lru_cache, partial
import random
from dataclasses import dataclass
from typing import List, Iterable, Any
from itertools import product
from math import prod
from copy import copy
from vidrial.kernels.mma_configurator import SMEM_LIMIT, MAX_THREADS_PER_BLOCK, evenly_divides, reasonable_atom_placements, dtype_to_bytes, static_shape, MMAAtom, safe_filter, REG_PER_SM, get_largest_atoms, SM_NUMBER, torch_dtype_to_c, MMAAtom_registry

logger = logging.getLogger(__name__)

@dataclass
class FlashConfig:
    Atom1: MMAAtom
    Atom2: MMAAtom
    MNKTileShape1: tuple[int, ...]
    NTile2: int
    MNKAtomPlacement1: tuple[int, ...]
    dtype: torch.dtype
    acc_dtype: torch.dtype
    MNDEP: tuple[int, ...] # (seq_q, seq_k, d, e, batch_size)
    smempipe1: int = 1
    smempipe2: int = 1
    regpipe1: int = 1
    regpipe2: int = 1
    use_ldsm1: bool = True
    use_ldsm2: bool = True
    swizzle1: int = 1
    swizzle2: int = 1
    q_in_reg: bool = True
    max_smem: int = SMEM_LIMIT
    max_threads: int = MAX_THREADS_PER_BLOCK

    @property
    def thread_count(self):
        return self.Atom1.thread_count * prod(self.MNKAtomPlacement1) # type: ignore

    @property
    def MNKAtomPlacement2(self):
        """Use second MMA's atom placement for filters like register estimate"""
        m_placement_1, n_placement_1, k_placement_1 = self.MNKAtomPlacement1
        total_atoms = self.thread_count // self.Atom1.thread_count
        atom1_shape = self.Atom1.MNK
        atom2_shape = self.Atom2.MNK
        m_placement_2 = m_placement_1 * atom1_shape[0] // atom2_shape[0]
        k_placement_2 = n_placement_1 * atom1_shape[1] // atom2_shape[1]
        n_placement_2 = total_atoms // (m_placement_2 * k_placement_2)
        assert n_placement_2 >= 1, "N_placement_2 must be at least 1"
        return (m_placement_2, n_placement_2, k_placement_2)
    
    @property
    def MNKP(self):
        """The MNKP of the second mma"""
        return (self.MNDEP[0], self.MNDEP[-2], self.MNDEP[1], self.MNDEP[-1])
    
    @property
    def MNKTileShape2(self):
        return (self.MNKTileShape1[0], self.NTile2, self.MNKTileShape1[1])
    
    @property
    def smem_estimate(self):
        M_tile_1, N_tile_1, d_tile_1 = self.MNKTileShape1
        Q_smem = M_tile_1 * self.MNDEP[2] * dtype_to_bytes(self.dtype)
        K_smem = N_tile_1 * d_tile_1 * dtype_to_bytes(self.dtype) * self.smempipe1
        V_smem = N_tile_1 * d_tile_1 * dtype_to_bytes(self.dtype) * self.smempipe2
        O_smem = M_tile_1 * N_tile_1 * dtype_to_bytes(self.dtype)
        l_smem = M_tile_1 * dtype_to_bytes(self.acc_dtype)
        if self.q_in_reg:
            return max(Q_smem, K_smem + V_smem, O_smem + l_smem)
        else:
            return max(Q_smem + K_smem + V_smem, O_smem + l_smem)

    @property
    def register_estimate(self):
        M_tile_1, N_tile_1, K_tile_1 = self.MNKTileShape1
        M_tile_2, N_tile_2, K_tile_2 = M_tile_1, self.NTile2, N_tile_1
        Q_reg = M_tile_1 * K_tile_1 // self.regpipe1 if not self.q_in_reg else M_tile_1 * self.MNDEP[2]
        K_reg = N_tile_1 * K_tile_1 // self.regpipe1
        S_reg = M_tile_2 * N_tile_1
        V_reg = N_tile_2 * K_tile_2 // self.regpipe2
        O_reg = M_tile_2 * N_tile_2
        if self.q_in_reg:
            return max(Q_reg + K_reg + S_reg, Q_reg + S_reg + V_reg + O_reg)
        else:
            return max(Q_reg + K_reg + S_reg, S_reg + V_reg + O_reg)
        
    @property
    def rest_m_1(self):
        return self.MNKTileShape1[0] // (self.Atom1.MNK[0] * self.MNKAtomPlacement1[0])
    
    @property
    def rest_n_1(self):
        return self.MNKTileShape1[1] // (self.Atom1.MNK[1] * self.MNKAtomPlacement1[1])
    
    @property
    def rest_k_1(self):
        return self.MNKTileShape1[2] // (self.Atom1.MNK[2] * self.MNKAtomPlacement1[2])
    
    @property
    def rest_m_2(self):
        return self.MNKTileShape2[0] // (self.Atom2.MNK[0] * self.MNKAtomPlacement2[0])
    
    @property
    def rest_n_2(self):
        return self.MNKTileShape2[1] // (self.Atom2.MNK[1] * self.MNKAtomPlacement2[1])
    
    @property
    def rest_k_2(self):
        return self.MNKTileShape2[2] // (self.Atom2.MNK[2] * self.MNKAtomPlacement2[2])

    @property
    def register_per_thread_estimate(self):
        rest_k_q = self.rest_k_1 // self.regpipe1 if not self.q_in_reg else self.MNDEP[2] // (self.Atom1.MNK[2] * self.MNKAtomPlacement1[2])
        Q_reg = self.Atom1.MNK[0] * self.Atom1.MNK[2] * self.rest_m_1 * rest_k_q // self.Atom1.thread_count
        K_reg = self.Atom1.MNK[1] * self.Atom1.MNK[2] * self.rest_n_1 * self.rest_k_1 // self.regpipe1 // self.Atom1.thread_count
        S_reg = self.Atom1.MNK[0] * self.Atom1.MNK[1] * self.rest_m_1 * self.rest_n_1 // self.Atom1.thread_count
        V_reg = self.Atom2.MNK[1] * self.Atom2.MNK[2] * self.rest_n_2 * self.rest_k_2 // self.regpipe2 // self.Atom2.thread_count
        O_reg = self.Atom2.MNK[0] * self.Atom2.MNK[1] * self.rest_m_2 * self.rest_n_2 // self.Atom2.thread_count
        if self.q_in_reg:
            return max(Q_reg + K_reg + S_reg, Q_reg + S_reg + V_reg + O_reg)
        else:
            return max(Q_reg + K_reg + S_reg, S_reg + V_reg + O_reg)

def get_valid_atoms(MNKTileShape, dtype, acc_dtype):
    """Returns all the atoms that evenly divide the tile shape and which have compatible types
    """
    Dtype = torch_dtype_to_c(dtype)
    AccDtype = torch_dtype_to_c(acc_dtype)
    valid_atoms = []
    for atom in [a for a in MMAAtom_registry.values() if not a.name.startswith('UniversalFMA')]:
        divides = evenly_divides(MNKTileShape, atom.MNK)
        valid_types = atom.A_type.value == Dtype and atom.B_type.value == Dtype and atom.C_type.value == AccDtype and atom.D_type.value == AccDtype
        if divides and valid_types:
            valid_atoms.append(atom)
    return valid_atoms

def reasonable_MNK_tile_shapes(M=None, N=None, K=None, MNK=None):
    """Generate a list of reasonable tile shapes for the MMA problem.
    """
    default_sizes = {1, 2, 4, 8, 16, 32, 64, 128}
    K_default_sizes = {1, 2, 4, 8, 16, 32, 64, 96, 128}
    M_sizes = default_sizes if M is None else {M,}
    N_sizes = default_sizes if N is None else {N,}
    K_sizes = K_default_sizes if K is None else {K,}
    for M in list(M_sizes):
        for N in list(N_sizes):
            for K in list(K_sizes):
                yield (M, N, K)

def second_mma_atom(atom, MNKTileShape, N_tile, dtype, acc_dtype):
    M_tile_1, N_tile_1, _ = MNKTileShape
    Tile = (M_tile_1, N_tile, N_tile_1)
    return get_valid_atoms(Tile, dtype, acc_dtype)

def initial_config_sweep(MNDEP, dtype, acc_dtype, M_tile=None, N_tile=None, K_tile=None, max_smem=SMEM_LIMIT, max_threads=MAX_THREADS_PER_BLOCK) -> list[FlashConfig]:
    configs = []
    MNK = MNDEP[:-2]
    E = MNDEP[-2]
    # Generate an initial large list of reasonable configs
    for MNKTileShape in reasonable_MNK_tile_shapes(M_tile, N_tile, K_tile, MNK):
        if not evenly_divides(MNK, MNKTileShape):
            continue
        for atom in get_valid_atoms(MNKTileShape, dtype, acc_dtype):
            for MNKAtomPlacement in reasonable_atom_placements(MNKTileShape, atom, max_threads):
                for NTile2 in {32, 64, 128}:
                    if E % NTile2 != 0:
                        continue
                    for atom2 in second_mma_atom(atom, MNKTileShape, NTile2, dtype, acc_dtype):
                        configs.append(FlashConfig(
                            Atom1=atom,
                            Atom2=atom2,
                            MNKTileShape1=MNKTileShape,
                            NTile2=NTile2,
                            MNKAtomPlacement1=MNKAtomPlacement,
                            dtype=dtype,
                            acc_dtype=acc_dtype,
                            MNDEP=MNDEP,
                        ))
    logger.info(f"Generated {len(configs)} initial mma configs")
    return configs


@safe_filter
def register_spillage_filter(spillage, configs):
    # We want to avoid register spillage both on per thread level and on per SM level
    configs = [c for c in configs if c.register_estimate <= REG_PER_SM and c.register_per_thread_estimate <= (255 + (spillage if c.q_in_reg else 0))]
    logger.info(f"Filtered to {len(configs)} flash configs for register spillage")
    return configs

@safe_filter
def register_underutil_filter(level, configs):
    # We want to avoid underutilizing the registers, if we are using < x% of
    # max registers, we filter out the configs that are underutilized
    top_usage = sorted([c.register_estimate for c in configs], reverse=True)[0]
    configs = [c for c in configs if c.register_estimate >= top_usage * level]
    logger.info(f"Filtered to {len(configs)} flash configs for register underutilization")
    return configs

@safe_filter
def smem_underutil_filter(level, configs):
    # We want to avoid underutilizing the smem, if we are using < x% of
    # max smem, we filter out the configs that are underutilized
    top_usage = sorted([c.smem_estimate for c in configs], reverse=True)[0]
    configs = [c for c in configs if c.smem_estimate >= top_usage * level]
    logger.info(f"Filtered to {len(configs)} flash configs for smem underutilization")
    return configs

@safe_filter
def regpipe_filter(regpipes, configs):
    # Filter out configs that use regpipe other than the ones we want to use
    configs = [c for c in configs if c.regpipe1 in regpipes and c.regpipe2 in regpipes]
    logger.info(f"Filtered to {len(configs)} flash configs for regpipe")
    return configs

@safe_filter
def smempipe_filter(smempipes, configs):
    # Filter out configs that use smempipe other than the ones we want to use
    configs = [c for c in configs if c.smempipe1 in smempipes and c.smempipe2 in smempipes]
    logger.info(f"Filtered to {len(configs)} flash configs for smempipe")
    return configs

@safe_filter
def warp_filter(num_warps, configs):
    # Filter out configs that use more warps than the number of warps we want to use
    configs = [c for c in configs if prod(c.MNKAtomPlacement1) <= num_warps]
    logger.info(f"Filtered to {len(configs)} flash configs for warp size")
    return configs

@safe_filter
def top_mtile_filter(top_k, configs):
    top_m_tiles = sorted([c.MNKTileShape1[0] for c in configs], reverse=True)[:top_k]
    configs = [c for c in configs if c.MNKTileShape1[0] in top_m_tiles]
    logger.info(f"Filtered to {len(configs)} flash configs for top mtile")
    return configs

@safe_filter
def thread_count_filter(configs):
    # If the candiate configs don't even use a full warp, pick the largerst one (relevant for small problems)
    thread_counts = [c.thread_count for c in configs]
    max_thread_count = max(thread_counts)
    if max_thread_count <= 32:
        configs = [c for c, t in zip(configs, thread_counts) if t == max_thread_count]

    # If there are configs that use multiples of 32 threads filter out the rest
    if any([t % 32 == 0 for t in thread_counts]):
        configs = [c for c, t in zip(configs, thread_counts) if t % 32 == 0]
    logger.info(f"Filtered to {len(configs)} flash configs due to thread count")
    return configs

@safe_filter
def atom_size_filter(configs):
    # Use the largest possible atom. Avoid benchmarking smaller instructions. TODO: verify this is a good idea
    largest_atom2s = get_largest_atoms([c.Atom2 for c in configs])
    configs = [c for c in configs if c.Atom2 in largest_atom2s]
    largest_atom1s = get_largest_atoms([c.Atom1 for c in configs])
    configs = [c for c in configs if c.Atom1 in largest_atom1s]
    logger.info(f"Filtered to {len(configs)} flash configs for largest atoms")
    return configs

@safe_filter
def saturation_filter(configs):
    MNKP = configs[0].MNKP
    # If the problem does not saturate the GPU, pick the config with the largest number of C tiles
    batch_size = MNKP[3] if isinstance(MNKP[3], int) else prod(MNKP[3])
    C_tile_nums = [ (MNKP[0]//c.MNKTileShape2[0]) * (MNKP[1]//c.MNKTileShape2[1]) * batch_size for c in configs]
    max_C_tile_num = max(C_tile_nums)
    if max_C_tile_num < SM_NUMBER:
        configs = [c for c, C_tile_num in zip(configs, C_tile_nums) if C_tile_num == max_C_tile_num]
    logger.info(f"Filtered to {len(configs)} flash configs for C tile count")
    return configs

@safe_filter
def arithmetic_intensity_filter(min_MN_tile_size, configs):
    # We want to pick as large a MNTile as possible to imporove arithmetic intensity
    MN_tile_sizes = [c.MNKTileShape2[0] * c.MNKTileShape2[1] for c in configs]
    if any([MN_tile_size > min_MN_tile_size for MN_tile_size in MN_tile_sizes]):
        configs = [c for c, MN_tile_size in zip(configs, MN_tile_sizes) if MN_tile_size >= min_MN_tile_size]
    logger.info(f"Filtered to {len(configs)} flash configs for large MN tile size")
    return configs

@safe_filter
def atom_placement_filter(configs):
    # we only pick atom placement that has k placement = 1
    configs = [c for c in configs if c.MNKAtomPlacement1[2] == 1 and c.MNKAtomPlacement2[2] == 1 and c.MNKAtomPlacement1[0] >= 1 and c.MNKAtomPlacement2[1] >= 1]
    logger.info(f"Filtered to {len(configs)} flash configs for atom placement")
    return configs

@safe_filter
def max_k_tile_filter(max_k_tile_num, configs):
    # we don't want there to be too many k tiles for the first mma
    configs = [c for c in configs if c.MNDEP[2] // c.MNKTileShape1[2] <= max_k_tile_num]
    logger.info(f"Filtered to {len(configs)} flash configs for max k tile number")
    return configs


def adjust_config(config: FlashConfig):
    """There're parameters that are being min/maxed inside the kernel, so we can prune them here"""
    config.smempipe2 = min(config.smempipe2, config.MNDEP[1] // config.MNKTileShape2[2])
    config.smempipe1 = config.smempipe2 * (config.MNDEP[2] // config.MNKTileShape1[2])
    config.regpipe1 = min(config.regpipe1, config.rest_k_1)
    config.regpipe2 = min(config.regpipe2, config.rest_k_2)
    return config


def deduplicate(configs: List[FlashConfig]) -> List[FlashConfig]:
    """Deduplicate configs using hash of the entire config."""
    seen = set()
    unique = []
    for config in configs:
        key = hash(str(config))
        if key not in seen:
            seen.add(key)
            unique.append(config)
    logger.info(f"Deduplicated {len(configs)} configs to {len(unique)} unique configs")
    return unique


def performance_expansion(configs: List[FlashConfig], smempipe=None, regpipe=None, use_ldsm=None, swizzle=None, q_in_reg=None):
    smempipes = [1, 2] if smempipe is None else (smempipe,) if not isinstance(smempipe, Iterable) else smempipe
    regpipes = [1] if regpipe is None else (regpipe,) if not isinstance(regpipe, Iterable) else regpipe
    use_ldsms = [True] if use_ldsm is None else (use_ldsm,) if not isinstance(use_ldsm, Iterable) else use_ldsm
    swizzles = [1] if swizzle is None else (swizzle,) if not isinstance(swizzle, Iterable) else swizzle
    q_in_regs = [True, False] if q_in_reg is None else (q_in_reg,) if not isinstance(q_in_reg, Iterable) else q_in_reg
    new_configs = []
    for config in configs:
        for smempipe, regpipe1, regpipe2, use_ldsm1, use_ldsm2, swizzle1, swizzle2, q_in_reg in product(smempipes, regpipes, regpipes, use_ldsms, use_ldsms, swizzles, swizzles, q_in_regs):
            c = copy(config)
            if (c.MNDEP[1] // c.MNKTileShape2[2]) < smempipe:
                continue
            c.smempipe1 = smempipe * (c.MNDEP[2] // c.MNKTileShape1[2])
            c.regpipe1 = regpipe1
            c.use_ldsm1 = use_ldsm1
            c.swizzle1 = swizzle1
            c.smempipe2 = smempipe
            c.regpipe2 = regpipe2
            c.use_ldsm2 = use_ldsm2
            c.swizzle2 = swizzle2
            c.q_in_reg = q_in_reg
            if c.smem_estimate > SMEM_LIMIT:
                continue
            new_configs.append(c)
    new_configs = [adjust_config(c) for c in new_configs]
    new_configs = deduplicate(new_configs)
    logger.info(f"Expanded to {len(new_configs)} configs with smempipe, regpipe, use_ldsm, swizzle for mma")
    return new_configs

def print_configs_table(configs: List[FlashConfig], max_rows: int = 50):
    """Print FlashConfig objects in a nicely formatted table."""
    if not configs:
        print("No configs to display.")
        return
    
    # Prepare the data for the table
    headers = [
        "ID", "Smem(KB)", "Reg/SM", "Reg/Th", "Th", "Atom1", "Atom2", "MNKTile1", "MNKPlace1", "NTile2", 
        "SMPipe1", "SMPipe2",
        "RegPipe1", "RegPipe2", "LDSM1", "LDSM2", "Swiz1", "Swiz2", "QinReg", 
    ]
    
    table_data = []
    for idx, c in enumerate(configs[:max_rows]):
        row = [
            idx,
            f"{c.smem_estimate / 1024:.1f}",
            c.register_estimate,
            c.register_per_thread_estimate,
            c.thread_count,
            c.Atom1.name,
            c.Atom2.name,
            f"{c.MNKTileShape1[0]}x{c.MNKTileShape1[1]}x{c.MNKTileShape1[2]}",
            f"{c.MNKAtomPlacement1[0]}x{c.MNKAtomPlacement1[1]}x{c.MNKAtomPlacement1[2]}",
            c.NTile2,
            c.smempipe1,
            c.smempipe2,
            c.regpipe1,
            c.regpipe2,
            "Y" if c.use_ldsm1 else "N",
            "Y" if c.use_ldsm2 else "N", 
            c.swizzle1,
            c.swizzle2,
            "Y" if c.q_in_reg else "N"
        ]
        table_data.append(row)
    
    # Print the table
    df = pd.DataFrame(table_data, columns=headers)
    pd.set_option('display.max_rows', None)
    pd.set_option('display.max_columns', None)
    pd.set_option('display.width', None)
    pd.set_option('display.max_colwidth', None)
    logger.debug(f"\nFlash Attention Configurations ({len(configs)} total, showing first {min(len(configs), max_rows)}):")
    logger.debug(f"\n{df}")
    
    # Print summary statistics
    logger.debug("\nSummary Statistics:")
    logger.debug(f"  Total configs: {len(configs)}")
    logger.debug(f"  Thread counts: {min(c.thread_count for c in configs)} - {max(c.thread_count for c in configs)}")
    logger.debug(f"  SMEM usage (KB): {min(c.smem_estimate / 1024 for c in configs):.1f} - {max(c.smem_estimate / 1024 for c in configs):.1f}")
    logger.debug(f"  Register/SM: {min(c.register_estimate for c in configs)} - {max(c.register_estimate for c in configs)}")
    logger.debug(f"  Register/Thread: {min(c.register_per_thread_estimate for c in configs)} - {max(c.register_per_thread_estimate for c in configs)}")

def render_configs(configs: List[FlashConfig]):
    new_configs = [
        {
            'Atom1': c.Atom1.name,
            'Atom2': c.Atom2.name,
            'MNKTileShape1': static_shape(c.MNKTileShape1),
            'MNKAtomPlacement1': static_shape(c.MNKAtomPlacement1),
            'NTile2': c.NTile2,
            **{k: v for k, v in c.__dict__.items() if k in ['smempipe1', 'regpipe1', 'use_ldsm1', 'swizzle1', 'smempipe2', 'regpipe2', 'use_ldsm2', 'swizzle2', 'q_in_reg']}
        }
        for c in configs
    ]
    logger.info(f"Filtered to {len(new_configs)} flash configs for kernel compatibility")
    return new_configs

def make_configurator(filter_funcs):
    @lru_cache(maxsize=None)
    def configurator(MNDEP, dtype, acc_dtype, M_tile=None, N_tile=None, K_tile=None, max_smem=SMEM_LIMIT, max_threads=MAX_THREADS_PER_BLOCK, smempipe=None, regpipe=None, use_ldsm=None, swizzle=None, q_in_reg=None, random_seed=None) -> list[dict[str, Any]]:
        configs = initial_config_sweep(MNDEP, dtype, acc_dtype, M_tile, N_tile, K_tile, max_smem, max_threads)
        configs = performance_expansion(configs, smempipe=smempipe, regpipe=regpipe, use_ldsm=use_ldsm, swizzle=swizzle, q_in_reg=q_in_reg)
        for filter_func in filter_funcs:
            configs = filter_func(configs)
        if random_seed is not None:
            random.seed(random_seed)
            random.shuffle(configs)
        
        # Print configs in a nice table format
        print_configs_table(configs)
        
        configs = render_configs(configs)
        return configs

    return configurator

no_filter_configurator = make_configurator([])

BASE_FILTERS = [
    atom_placement_filter,
    thread_count_filter, # For small problems, we want to use the maximum number of threads
    atom_size_filter, # Use the largest possible atom. Avoid benchmarking smaller instructions. TODO: verify this is a good idea
    partial(register_spillage_filter, 0), # Avoid register spillage
    saturation_filter, # If the problem does not saturate the GPU, pick the config with the largest number of C tiles
    partial(arithmetic_intensity_filter, 2048), # We want to pick as large a MNTile as possible to imporove arithmetic intensity
]

base_configurator = make_configurator(BASE_FILTERS)

ADVANCED_FILTERS = [
    atom_placement_filter,
    thread_count_filter,
    atom_size_filter,
    # partial(top_mtile_filter, 2), # We want to pick the top 2 mtiles
    partial(register_spillage_filter, 20), # Avoid register spillage
    saturation_filter,
    # partial(max_k_tile_filter, 2), # we don't want there to be too many k tiles for the first mma
    partial(arithmetic_intensity_filter, 2048),
    partial(smem_underutil_filter, 0.3), # Avoid underutilizing the smem relative to maximum smem usage
    partial(register_underutil_filter, 0.3), # Avoid underutilizing the registers relative to maximum register usage
    partial(warp_filter, 4), # Avoid using more warps than the number of warps we want to use
    partial(smempipe_filter, [1, 2, 3, 4]), # Smempipe 1 and 2 and 3 and 4 are the only ones we want to use
    partial(regpipe_filter, [1, 2]), # Regpipe 1 is probably good enough
]

advanced_configurator = make_configurator(ADVANCED_FILTERS)


import pytest
import yaml
from vidrial.jit.static.types import Shape
from vidrial.jit.timingcache import ConfigTimingCache
import torch

def test_basic_cache_operations(tmp_path):
    # Simple hash function for testing
    hash_fn = lambda x: str(sorted(x.items()))
    
    cache = ConfigTimingCache("test_fn", hash_fn, root=str(tmp_path))
    
    # Test insertion and retrieval
    arg = {"param1": 1, "param2": 2}
    timings = [
        {"config": {"tile": 32}, "time": 0.5, "std": 0.1, "n": 10},
        {"config": {"tile": 64}, "time": 0.3, "std": 0.05, "n": 20}
    ]
    
    cache.update(arg, timings)
    
    # Test __contains__
    assert arg in cache
    
    # Test __getitem__ - should return best performing config
    best_config = cache[arg]
    assert best_config == {"tile": 64}  # The one with time 0.3
    
    # Test __len__
    assert len(cache) == 1

def test_sorting_behavior(tmp_path):
    hash_fn = lambda x: str(sorted(x.items()))
    cache = ConfigTimingCache("test_fn", hash_fn, root=str(tmp_path))
    
    arg = {"param": 1}
    
    # Add timings in non-sorted order
    timings = [
        {"config": {"tile": 32}, "time": 0.5},
        {"config": {"tile": 16}, "time": 0.2},
        {"config": {"tile": 64}, "time": 0.3}
    ]
    
    cache.update(arg, timings)
    
    # Verify best config is returned (should be tile=16 with time=0.2)
    assert cache[arg] == {"tile": 16}
    
    # Add another config that hashes to same key
    config2 = {"param": 1}  # Same hash as config
    timings2 = [
        {"config": {"tile": 8}, "time": 0.1}
    ]
    
    cache.update(config2, timings2)
    
    # Verify entries are sorted by best timing
    key = hash_fn(arg)
    entries = cache._cache[key]
    assert entries[0]["timings"][0]["time"] == 0.1  # Best timing should be first
    assert entries[1]["timings"][0]["time"] == 0.2  # Second best timing next

def test_size_limiting(tmp_path):
    hash_fn = lambda x: str(x["id"])
    cache = ConfigTimingCache("test_fn", hash_fn, root=str(tmp_path), size=2)
    
    # Add 3 configs - should only keep 2
    for i in range(3):
        config = {"id": i}
        timings = [{"config": {"tile": 32}, "time": float(i)}]
        cache.update(config, timings)
    
    assert len(cache) == 2
    # Should keep the ones with better timings
    assert {"id": 0} in cache
    assert {"id": 1} in cache
    assert {"id": 2} not in cache

def test_file_persistence(tmp_path):
    hash_fn = lambda x: str(sorted(x.items()))
    
    # Create and populate cache
    cache1 = ConfigTimingCache("test_fn", hash_fn, root=str(tmp_path))
    arg = {"param": 1}
    timings = [{"config": {"tile": 32}, "time": 0.5, "std": 0.1, "n": 10}]
    cache1.update(arg, timings)
    
    # Create new cache instance - should load from file
    cache2 = ConfigTimingCache("test_fn", hash_fn, root=str(tmp_path))
    assert arg in cache2
    assert cache2[arg] == {"tile": 32}
    
    # Test remove functionality
    cache2.remove()
    cache3 = ConfigTimingCache("test_fn", hash_fn, root=str(tmp_path))
    assert len(cache3) == 0

def test_strict_verification(tmp_path):
    hash_fn = lambda x: str(sorted(x.items()))
    cache = ConfigTimingCache("test_fn", hash_fn, root=str(tmp_path))
    
    arg = {"param": 1}
    timings = [
        {"config": {"tile": 32}, "time": 0.5, "std": 0.1, "n": 10},
        {"config": {"tile": 16}, "time": 0.2, "std": 0.05, "n": 20}
    ]
    cache.update(arg, timings)
    
    # Manually corrupt the cache file to test verification
    key = hash_fn(arg)
    with open(cache.location, 'r') as f:
        data = yaml.safe_load(f)
    
    # Corrupt the sorting
    data[key][0]["timings"].reverse()
    
    with open(cache.location, 'w') as f:
        yaml.safe_dump(data, f)
    
    # Loading with strict=True should raise error
    with pytest.raises(ValueError):
        ConfigTimingCache("test_fn", hash_fn, root=str(tmp_path), strict=True)
    
    # Loading with strict=False should work
    cache2 = ConfigTimingCache("test_fn", hash_fn, root=str(tmp_path), strict=False)
    assert arg in cache2

def test_storing_vidrial_types(tmp_path):
    x = torch.randn(10, 10)
    hash_fn = lambda arg: str(arg['x'].shape)
    arg = {'x': x}
    cache = ConfigTimingCache("test_fn", hash_fn, root=str(tmp_path))
    tile_shape = Shape(32, 32)
    cache.update(arg, [{"config": {"tile": tile_shape}, "time": 0.5, "std": 0.1, "n": 10}])
   
    assert arg in cache
    assert cache[arg] == {"tile": tile_shape}

    # test loading from file
    cache2 = ConfigTimingCache("test_fn", hash_fn, root=str(tmp_path))
    assert arg in cache2
    assert cache2[arg] == {"tile": tile_shape}


import torch as th
from math import sqrt
from vidrial.jit.static.types import Shape, Layout, Int
from vidrial.kernels.sympow.dimensions import problem_dimensions, op_output_shape, SympowCoords
from vidrial.kernels.sympow.binding import binding_autotuned
from vidrial.kernels.sympow_bwd.op import op as sympow_bwd
from vidrial.py_utils.common import tprod


# ---------------------- Op --------------------------------

@th.library.custom_op("vidrial::sympow", mutates_args=())
def op(X: th.Tensor, power: int, d_tile: int, duplicate_correction: bool = True) -> th.Tensor:
    """ Computes the tiled symmetric power of X.
    Dimensions:
        d the feature dimension being powered.
        bs=[b0, b1, ...] the batch dimensions.
        D the expanded dimension. Given by sympow_dim(d, power, d_tile)
    Args:
        X: Input tensor of shape `(b0, b1, ..., d)`.
        power: Power of the symmetric power.
        d_tile: Tile size for the operation.
        duplicate_correction: Whether to correct for duplicate elements.
    Returns:
        Z: Output tensor of shape `(b0, b1, ..., nt, dt, dt, ...,)`. Where dt = d_tile and
           nt is the number of tiles.
    """
    expand_batch = True if X.dim() == 1 else False # expand a batch dimension if there is none
    if expand_batch: X = X.unsqueeze(0)
    assert X.device.type == "cuda", "X must be on GPU"
    Z = th.empty(op_output_shape(X.shape, power, d_tile), device=X.device, dtype=X.dtype)
    binding_autotuned(Z, X, power, d_tile, duplicate_correction)
    if expand_batch: Z = Z[0]
    return Z

@op.register_fake
def op_fake(X: th.Tensor, power: int, d_tile: int, duplicate_correction: bool = True):
    return th.empty(op_output_shape(X.shape, power, d_tile), device=X.device, dtype=X.dtype)


# ---------------------- Autograd --------------------------------

def setup_context(ctx, inputs, output):
    X, power, d_tile, duplicate_correction = inputs
    ctx.save_for_backward(X)
    ctx.power = power
    ctx.d_tile = d_tile
    ctx.duplicate_correction = duplicate_correction
def op_backward(ctx, dZ):
    X, = ctx.saved_tensors
    dX = sympow_bwd(X, dZ, ctx.power, ctx.d_tile, ctx.duplicate_correction)
    return dX, None, None, None
th.library.register_autograd(
    "vidrial::sympow", op_backward, setup_context=setup_context
)


# ---------------------- Reference Op  --------------------------------

def op_reference(X: th.Tensor, power: int, d_tile: int, duplicate_correction: bool = True) -> th.Tensor:
    expand_batch = True if X.dim() == 1 else False
    if expand_batch: X = X.unsqueeze(0)
    Z = th.empty(op_output_shape(X.shape, power, d_tile), device=X.device, dtype=X.dtype)
    bs, d, D = problem_dimensions(Z.shape, X.shape, power, d_tile)
    batch_slice = (slice(None),) * len(bs)
    x_tiled = X.reshape(bs + (-1, d_tile))
    for idx, seq, dup in SympowCoords(d // d_tile, power):
        z_tile = Z[batch_slice + (idx,)]
        x_tiles = [x_tiled[batch_slice + (seq[i],)] for i in range(power)]
        z_tile[:] = tprod(*x_tiles)
        if duplicate_correction: z_tile[:] *= sqrt(dup)
    if expand_batch: Z = Z[0]
    return Z



from dataclasses import dataclass
import math
from typing import Any
import torch
from vidrial.jit.static.types import Shape, Layout, Int
from vidrial.jit.static.util import layout_from_shape_stride, torch_dtype_to_c
from vidrial.jit.timingcache import ConfigTimingCache
from vidrial.jit.binding import make_binding
from vidrial.jit.jit import jit, render
from vidrial.kernels.flash.configurator import base_configurator, advanced_configurator


# ------------------- Source Code -------------------

@dataclass
class SourceCode:
    T: str
    LT: str
    Atom1: str
    Atom2: str
    MNKTileShape1: Shape
    NTile2: Int
    MNKAtomPlacement1: Shape
    GQSlab: Layout
    GKSlab: Layout
    GVSlab: Layout
    GOSlab: Layout
    GLSlab: Layout
    smempipe1: int
    smempipe2: int
    regpipe1: int
    regpipe2: int
    use_ldsm1: bool
    use_ldsm2: bool
    swizzle1: int
    swizzle2: int
    q_in_reg: bool
    @property
    def template(self):
        return """
#include <cuda.h>
#include <cuda_runtime.h>
#include <iostream>
#include "cute/tensor.hpp"
#include "kernels/flash/kernel.cuh"

using namespace cute;
using namespace vidrial;

extern "C" int launch(void* __raw_Q, void* __raw_K, void* __raw_V, void* __raw_O, void* __raw_l, float softmax_scale) {
    using T = {T};
    using LT = {LT};
    using Atom1 = MMA_Atom<{Atom1}>;
    using Atom2 = MMA_Atom<{Atom2}>;
    using MNKTileShape_1 = decltype(static_tree_cast<int64_t>({MNKTileShape1}{}));
    using NTile2 = decltype(static_tree_cast<int64_t>({NTile2}{}));
    using MNKAtomPlacement_1 = decltype(static_tree_cast<int64_t>({MNKAtomPlacement1}{}));
    using GQSlab = decltype(static_tree_cast<int64_t>({GQSlab}{}));
    using GKSlab = decltype(static_tree_cast<int64_t>({GKSlab}{}));
    using GVSlab = decltype(static_tree_cast<int64_t>({GVSlab}{}));
    using GOSlab = decltype(static_tree_cast<int64_t>({GOSlab}{}));
    using GLSlab = decltype(static_tree_cast<int64_t>({GLSlab}{}));
    using PerfCfg1 = FlashPerfCfg<{smempipe1}, {regpipe1}, {use_ldsm1}, {swizzle1}, {q_in_reg}>;
    using PerfCfg2 = PerfCfg<{smempipe2}, {regpipe2}, {use_ldsm2}, {swizzle2}>;
    auto cfg = make_FlashKernelCfg<T, LT, Atom1, MNKTileShape_1, MNKAtomPlacement_1, Atom2, NTile2, GQSlab, GKSlab, GVSlab, GOSlab, GLSlab, PerfCfg1, PerfCfg2>();
    auto Q = reinterpret_cast<T*>(__raw_Q);
    auto K = reinterpret_cast<T*>(__raw_K);
    auto V = reinterpret_cast<T*>(__raw_V);
    auto O = reinterpret_cast<T*>(__raw_O);
    auto l = reinterpret_cast<LT*>(__raw_l);
    return launch_flash_attention_kernel(cfg, Q, K, V, O, l, softmax_scale);
}"""
    def __str__(self):
        return render(self.template, self.__dict__)


# ------------------- Binding -------------------

@dataclass
class BindingCfg:
    batches: tuple[int, ...]
    M: int
    N: int
    D: int
    E: int
    Q_shape: tuple[int, ...]
    K_shape: tuple[int, ...]
    V_shape: tuple[int, ...]
    O_shape: tuple[int, ...]
    l_shape: tuple[int, ...]
    Q_stride: tuple[int, ...]
    K_stride: tuple[int, ...]
    V_stride: tuple[int, ...]
    O_stride: tuple[int, ...]
    l_stride: tuple[int, ...]
    Atom1: str
    Atom2: str
    MNKTileShape1: tuple[int, ...]
    NTile2: int
    MNKAtomPlacement1: tuple[int, ...]
    smempipe1: int
    smempipe2: int
    regpipe1: int
    regpipe2: int
    use_ldsm1: bool
    use_ldsm2: bool
    swizzle1: int
    swizzle2: int
    q_in_reg: bool
    dtype: torch.dtype
    acc_dtype: torch.dtype
    @classmethod
    def from_args(cls, Q, K, V, O, l, Atom1, Atom2, MNKTileShape1, NTile2, MNKAtomPlacement1, smempipe1, smempipe2, regpipe1, regpipe2, use_ldsm1, use_ldsm2, swizzle1, swizzle2, q_in_reg):
        assert Q.device.type == K.device.type == V.device.type == O.device.type == l.device.type == 'cuda', f"Invalid {Q.device}, {K.device}, {V.device}, {O.device}, {l.device}."
        assert Q.dtype == K.dtype == V.dtype == O.dtype, f"Invalid {Q.dtype}, {K.dtype}, {V.dtype}, {O.dtype},  Q, K, V, O must have the same dtype"
        batches, M, N, D, E = Q.shape[:-2], Q.shape[-2], K.shape[-2], K.shape[-1], V.shape[-1]
        assert K.shape[:-2] == batches, f"Invalid {K.shape}, K must be of shape ({', '.join(str(b) for b in batches)}, {K.shape[-2]}, {K.shape[-1]})"
        assert V.shape[:-2] == batches, f"Invalid {V.shape}, V must be of shape ({', '.join(str(b) for b in batches)}, {V.shape[-2]}, {V.shape[-1]})"
        assert O.shape == (*batches, M, E), f"Invalid {O.shape}, O must be of shape ({', '.join(str(b) for b in batches)}, {M}, {E})"
        assert l.shape == (*batches, M), f"Invalid {l.shape}, l must be of shape ({', '.join(str(b) for b in batches)}, {M})"

        return cls(batches, M, N, D, E, Q.shape, K.shape, V.shape, O.shape, l.shape, Q.stride(), K.stride(), V.stride(), O.stride(), l.stride(), Atom1, Atom2, MNKTileShape1, NTile2, MNKAtomPlacement1, smempipe1, smempipe2, regpipe1, regpipe2, use_ldsm1, use_ldsm2, swizzle1, swizzle2, q_in_reg, Q.dtype, l.dtype)

    @property
    def source(self):
        """ Conversion between the different variable name and shape conventions
                Python                        c++
            Q_shape=[P,M1,K1]      ->      GQSlabShape=[M1,K1,P]
            K_shape=[P,N1,K1]      ->      GKSlabShape=[N1,K1,P]
            V_shape=[P,K2,N2]      ->      GVSlabShape=[N2,K2,P]
            O_shape=[P,M2,N2]      ->      GO_slabShape=[M2,N2,P]
            l_shape=[P,M2]      ->      GL_slabShape=[M2,P]
        """
        batch_inds = tuple(range(len(self.batches)))
        GQslab = layout_from_shape_stride(self.Q_shape, self.Q_stride, (-2, -1, batch_inds))
        GKslab = layout_from_shape_stride(self.K_shape, self.K_stride, (-2, -1, batch_inds))
        GVslab = layout_from_shape_stride(self.V_shape, self.V_stride, (-1, -2, batch_inds))
        GOslab = layout_from_shape_stride(self.O_shape, self.O_stride, (-2, -1, batch_inds))
        GLslab = layout_from_shape_stride(self.l_shape, self.l_stride, (-1, batch_inds))
        return SourceCode(
            T=torch_dtype_to_c(self.dtype),
            LT=torch_dtype_to_c(self.acc_dtype),
            Atom1=self.Atom1,
            Atom2=self.Atom2,
            MNKTileShape1=Shape(*[Int(x) for x in self.MNKTileShape1]),
            NTile2=Int(self.NTile2),
            MNKAtomPlacement1=Shape(*[Int(x) for x in self.MNKAtomPlacement1]),
            GQSlab=GQslab,
            GKSlab=GKslab,
            GVSlab=GVslab,
            GOSlab=GOslab,
            GLSlab=GLslab,
            smempipe1=self.smempipe1,
            smempipe2=self.smempipe2,
            regpipe1=self.regpipe1,
            regpipe2=self.regpipe2,
            use_ldsm1=self.use_ldsm1,
            use_ldsm2=self.use_ldsm2,
            swizzle1=self.swizzle1,
            swizzle2=self.swizzle2,
            q_in_reg=self.q_in_reg,
        )

# ---------------------- Autotune --------------------------------

def make_mma_configurator(smempipe=None, regpipe=None, use_ldsm=None, swizzle=None):
    def mma_configurator(args: dict) -> list[dict[str, Any]]:
        Q, K, V = args['Q'], args['K'], args['V']
        batches, M, N, D, E = Q.shape[:-2], Q.shape[-2], K.shape[-2], K.shape[-1], V.shape[-1]
        return advanced_configurator(MNDEP=(M, N, D, E, batches), dtype=Q.dtype, acc_dtype=torch.float32, smempipe=smempipe, regpipe=regpipe, use_ldsm=use_ldsm, swizzle=swizzle, random_seed=42)
    return mma_configurator

def hash_fn(args: dict) -> str:
    Q, K, V, O, l = args['Q'], args['K'], args['V'], args['O'], args['l']
    batches, M, N, D, E = Q.shape[:-2], Q.shape[-2], K.shape[-2], K.shape[-1], V.shape[-1]
    Q_major = 'K_major' if Q.stride(-1) == 1 else 'M_major'
    K_major = 'N_major' if K.stride(-1) == 1 else 'K_major'
    V_major = 'N_major' if V.stride(-1) == 1 else 'K_major'
    P = math.prod(batches)
    return f"{Q.dtype}-{K.dtype}-{V.dtype}-{O.dtype}-{l.dtype}-M_{M}-N_{N}-D_{D}-E_{E}-P_{P}-{Q_major}-{K_major}-{V_major}"

@make_binding(cache=ConfigTimingCache('flash_attn', hash_fn),
          sweep=make_mma_configurator(smempipe=(1,2,3), regpipe=(1,2,4), use_ldsm=True, swizzle=1))
def binding(Q, K, V, O, l, softmax_scale, Atom1, Atom2, MNKTileShape1, NTile2, MNKAtomPlacement1, smempipe1, smempipe2, regpipe1, regpipe2, use_ldsm1, use_ldsm2, swizzle1, swizzle2, q_in_reg):
    binding_cfg = BindingCfg.from_args(Q, K, V, O, l, Atom1, Atom2, MNKTileShape1, NTile2, MNKAtomPlacement1, smempipe1, smempipe2, regpipe1, regpipe2, use_ldsm1, use_ldsm2, swizzle1, swizzle2, q_in_reg)
    jit(name = "flash_attn",
        code = str(binding_cfg.source)
    )(Q, K, V, O, l, softmax_scale)



# credit: 
import copy
import ctypes
import torch
from typing import Any, Dict, Iterable, Tuple


# Type map from torch dtype to cutlass dtype
dtype_map = {
    torch.float16: 'cutlass::half_t',
    torch.float32: 'float',
    torch.float64: 'double',
    torch.bfloat16: 'cutlass::bfloat16_t',
}

# Name map for Python `eval`
typename_map: Dict[Any, str] = {
    **{t: t.__name__ for t in (bool, int, float)},
    torch.int: 'torch.int',
    torch.float: 'torch.float',
    torch.float16: 'torch.float16',
    torch.bfloat16: 'torch.bfloat16',
    torch.float32: 'torch.float32',
    torch.float64: 'torch.float64',
    torch.float8_e4m3fn: 'torch.float8_e4m3fn',
    torch.cuda.Stream: 'torch.cuda.Stream',
}

# `ctype` map for Python casting
ctype_map: Dict[Any, Any] = {
    **{t: getattr(ctypes, f'c_{t.__name__}') for t in (bool, int, float)},
    **{t: ctypes.c_void_p for t in (torch.int, torch.float, torch.float32, torch.float64, torch.bfloat16, torch.float8_e4m3fn, torch.cuda.Stream)},
}

# Type map for both Python API and source code usages
genc_map = {
    bool: ('bool', 'bool'),
    int: ('int', 'int'),
    float: ('float', 'float'),
    torch.int: ('void*', 'int*'),
    torch.float: ('void*', 'float*'),
    torch.float32: ('void*', 'float*'),
    torch.float64: ('void*', 'double*'),
    torch.float16: ('void*', 'cutlass::half_t*'),
    torch.bfloat16: ('void*', 'cutlass::bfloat16_t*'),
    torch.float8_e4m3fn: ('void*', 'cutlass::float_e4m3_t*'),
    torch.cuda.Stream: ('void*', 'cudaStream_t'),
}

def map_ctype(value: Any) -> Any:
    if hasattr(value, 'data_ptr'):
        if value.dtype == torch.int:
            return ctypes.c_void_p(value.data_ptr())
        elif value.dtype == torch.float:
            return ctypes.c_void_p(value.data_ptr())
        elif value.dtype == torch.float32:
            return ctypes.c_void_p(value.data_ptr())
        elif value.dtype == torch.float64:
            return ctypes.c_void_p(value.data_ptr())
        elif value.dtype == torch.bfloat16:
            return ctypes.c_void_p(value.data_ptr())
        elif value.dtype == torch.float16:
            return ctypes.c_void_p(value.data_ptr())
        elif value.dtype == torch.float8_e4m3fn:
            return ctypes.c_void_p(value.data_ptr())
        else:
            return ctypes.c_void_p(value.data_ptr())

    if hasattr(value, 'cuda_stream'):
        return ctypes.c_void_p(value.cuda_stream)

    if isinstance(value, bool):
        return ctypes.c_bool(value)
    elif isinstance(value, int):
        return ctypes.c_int(value)
    elif isinstance(value, float):
        return ctypes.c_float(value)

    return ctype_map[type(value)](value)



import os
os.environ['JIT_PACKAGE_MODE'] = '1'

from vidrial.jit.timer import timeit
from flash_attn.flash_attn_interface import flash_attn_func
from vidrial.jit.settings import settings, PickBest
from vidrial.kernels.flash.op import op as flash_vidrial
from vidrial.py_utils.test_utils import diff
import torch
import pandas as pd
import logging
logging.basicConfig(level=logging.INFO)

def create_input(b, tq, tk, h, d, e):
    Q = torch.randn(b, tq, h, d, device='cuda', dtype=torch.bfloat16)
    K = torch.ones(b, tk, h, d, device='cuda', dtype=torch.bfloat16)
    V = torch.randn(b, tk, h, e, device='cuda', dtype=torch.bfloat16)
    O = torch.empty(b, tq, h, e, device='cuda', dtype=torch.bfloat16)
    l = torch.empty(b, tq, h, device='cuda', dtype=torch.float32)
    return Q, K, V, O, l

def compare():
    with settings.set(policy=PickBest, max_configs=512, plot_timings=False):
        result = []
        for b, tq, tk, h, d, e in [(1, 1024, 1024, 16, 64, 64),
                                   (1, 1024, 1024, 16, 32, 32)]:
            Q, K, V, O, l = create_input(b, tq, tk, h, d, e)
            O_flash = flash_attn_func(Q, K, V, causal=True)
            O_vidrial, l_vidrial = flash_vidrial(Q, K, V)

            cuda_avg, cuda_std = timeit(flash_vidrial, Q, K, V, num1=10, num2=30)
            flash_avg, flash_std = timeit(flash_attn_func, Q, K, V, causal=True, num1=10, num2=30)
            max_diff = torch.max(torch.abs(O_vidrial - O_flash)).item()
            result.append({
                'b': b,
                'tq': tq,
                'tk': tk,
                'h': h,
                'd': d,
                'e': e,
                'cuda_avg': cuda_avg,
                'cuda_std': cuda_std,
                'flash_avg': flash_avg,
                'flash_std': flash_std,
                'speedup': flash_avg / cuda_avg,
                'max_diff': max_diff
            })
        return pd.DataFrame(result)

if __name__ == "__main__":
    import logging
    logging.basicConfig(level=logging.DEBUG)
    df = compare()
    print(df)


import torch
from vidrial.jit.compiler import compile


def test_add_one(get_add_one_kernel_code):
    X = torch.randn(64, 64, device="cuda")
    Y = torch.empty_like(X)
    code = get_add_one_kernel_code(X, Y)
    runtime = compile(code)
    runtime(X, Y)
    torch.testing.assert_close(X + 1, Y)



from typing import Any, Union, Iterator, Iterable
from torch.utils._pytree import register_pytree_node
from vidrial.jit.static.int import Int, BaseType
from functools import reduce

class Tuple(BaseType):
    """Represents a hierarchical integer tuple type. A tuple can be constructed
    from a (mixed) list of ints, tuples, or Ints, but once constructed,
    all elements are either Ints or Tuples."""
    elements: tuple[Union[Int, 'Tuple', int], ...]
    name: str = "tuple"

    def __init__(self, *elements: Union[int, 'Tuple', Int]):
        assert all(isinstance(elem, (Int, Tuple, int, tuple)) for elem in elements), f"elements={elements} should be a list of Ints or Tuples"
        _elements = []
        for elem in elements:
            if isinstance(elem, tuple):
                _elements.append(self.__class__(*elem))
            else:
                _elements.append(elem)
        self.elements = tuple(_elements)

    def __getitem__(self, idx: Union[int, slice]) -> Union['Tuple', Int, int]:
        if isinstance(idx, int):
            return self.elements[idx]
        elif isinstance(idx, slice):
            return self.__class__(*self.elements[idx])
        else:
            raise ValueError(f"Invalid index type: {type(idx)}")

    def __iter__(self) -> Iterator[Union['Tuple', Int, int]]:
        return iter(self.elements)
    
    def to_cpp(self, enclosed: bool = False) -> str:
        elements_str = ", ".join(elem.to_cpp(True) if isinstance(elem, BaseType) else str(elem) for elem in self.elements)
        if enclosed and len(self.elements) == 1:
            return elements_str
        return f"{self.name}<{elements_str}>"
    
    def size(self) -> int:
        return reduce(lambda x, y: x * (y.size() if isinstance(y, BaseType) else y), self.elements, 1)
    
    def rank(self) -> int:
        return len(self.elements)
    
    def depth(self) -> int:
        return max(elem.depth() + 1 if isinstance(elem, self.__class__) else 1 for elem in self.elements)
    
    def modes(self) -> int:
        return sum(elem.modes() if isinstance(elem, self.__class__) else 1 for elem in self.elements)
    
    def to_tuple(self) -> tuple[Any, ...]:
        """Convert a Tuple to a python tuple containing python ints or tuples of ints."""
        return tuple(elem.to_tuple() if isinstance(elem, self.__class__) else elem for elem in self.elements)
    
    def to_list(self) -> list[Any]:
        """Convert a Tuple to a python list containing python ints or tuples of ints."""
        return [elem.to_list() if isinstance(elem, self.__class__) else elem for elem in self.elements]

    def __hash__(self) -> int:
        return hash(self.to_tuple())

    def __str__(self) -> str:
        if len(self.elements) == 1:
            if isinstance(self.elements[0], Int):
                return str(self.elements[0])
            else:
                return f"({','.join(str(elem) for elem in self.elements)})"
        else:
            return f"({','.join(str(elem) for elem in self.elements)})"
        
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({', '.join(repr(elem) for elem in self.elements)})"
    
    def __len__(self) -> int:
        return len(self.elements)
    
    def __contains__(self, item: Any) -> bool:
        return item in self.elements
    
    def __eq__(self, other: 'Tuple') -> bool:
        if not isinstance(other, self.__class__):
            return False
        if len(self.elements) != len(other.elements):
            return False
        return all(elem == other.elements[i] for i, elem in enumerate(self.elements)) # type: ignore
    
    def __reduce__(self):
        """Make Tuple picklable by specifying how to reconstruct it."""
        return (type(self), self.elements)


def flatten(tup: Tuple) -> Tuple:
    """Recursively flatten a tuple of tup."""
    flattened = []
    cls = type(tup)
    for item in tup:
        if isinstance(item, Tuple):
            flattened.extend(flatten(item))
        else:
            flattened.append(item)
    return cls(*flattened)


def unflatten(flat: Iterable[Any], target: Tuple) -> Tuple:
    """Unflatten a flat tuple into a the structure of the target tuple."""
    flat = list(flat)
    assert len(flat) == len(flatten(target)), f"flat={flat} and target={target} must have the same number of elements"
    res = []
    cls = type(target)
    for item in target:
        if isinstance(item, (Int, int)):
            res.append(flat[0])
            flat = flat[1:] # type: ignore
        elif isinstance(item, Tuple):
            res.append(unflatten(flat[:item.modes()], item)) # type: ignore
            flat = flat[item.modes():] # type: ignore
    return cls(*res)
        

register_pytree_node(Tuple, lambda t: (list(flatten(t)), t), unflatten)



import torch as th
import pytest
import time
from vidrial.kernels.add_one.add_one import add_one_kernel

# To run with all the JIT logs: export JIT_VERBOSE=1; uv run pytest -s

def test_add_one_basic():
    """Test basic functionality of add_one."""
    V = th.ones(32, 32, dtype=th.float32, device="cuda")
    V_ref = V + 1
    
    add_one_kernel(V)
    
    assert V.shape == (32, 32)
    assert V.dtype == th.float32
    assert V.device.type == "cuda"
    assert th.allclose(V, V_ref, rtol=1e-5, atol=1e-5)



import torch as th
from vidrial.kernels.sympow.dimensions import sympow_dim

def dimensions(A_shape, B_shape, c_shape, power, d_tile):
    """ Checks that the tensors satisfy the shape assumptions and return the problem dimensions """
    batches, M, K, N, d =  tuple(A_shape[:-2]), A_shape[-2], A_shape[-1], B_shape[-1], c_shape[-2]
    D = sympow_dim(d, power, d_tile)
    assert M == D, "since expand_dim=-1, "
    assert A_shape == batches + (M, K)
    assert B_shape == batches + (K, N)
    assert c_shape == batches + (d, N)
    return batches, M, K, N, d

def output_shape(A_shape, B_shape, c_shape, power, d_tile):
    dimensions(A_shape, B_shape, c_shape, power, d_tile) # checks that the dimensions are valid
    return c_shape
 
def canonical_inputs(A, B, c, c_dot, expand_dim):
    """ Make all the arguments 3D and ensure expand_dim indexes with negative values """
    assert expand_dim == -2, f"{expand_dim} not implemented yet. Current cuda kernel only supports expanding -2"
    assert A.dtype == B.dtype == c.dtype == c_dot.dtype, f"Invalid {A.dtype}, {B.dtype}, {c.dtype}, {c_dot.dtype}. Kernel currently assumes all input tensors have the same dtype"
    assert A.ndim == B.ndim == c.ndim == c_dot.ndim, f"Invalid {A.ndim}, {B.ndim}, {c.ndim}, {c_dot.ndim}. Kernel currently assumes all input tensors have the same number of dimensions"
    if A.ndim == 2: #  there must be at least 1 batch dimension
        assert B.ndim == c.ndim == c_dot.ndim == 2, f"Invalid {A.ndim}, {B.ndim}, {c.ndim}, {c_dot.ndim}. Kernel currently only supports 2D input tensors"
        A, B, c, c_dot = map(lambda t: th.unsqueeze(t, 0), (A, B, c, c_dot))
        expand_dim += 1
    expand_dim = expand_dim%A.ndim-A.ndim  # make expand_dim negative
    return A, B, c, c_dot, expand_dim



import torch
import logging

logger = logging.getLogger(__name__)

# 24 permutations of strides up to 4D
PERMUTATIONS_OF_STRIDES_UP_TO_4D = [
    (0, 1, 2, 3),
    # (0, 1, 3, 2),
    # (0, 2, 1, 3),
    # (0, 2, 3, 1),
    # (0, 3, 1, 2),
    # (0, 3, 2, 1),
    # (1, 0, 2, 3),
    # (1, 0, 3, 2),
    # (1, 2, 0, 3),
    (1, 2, 3, 0),
    # (1, 3, 0, 2),
    # (1, 3, 2, 0),
    # (2, 0, 1, 3),
    # (2, 0, 3, 1),
    # (2, 1, 0, 3),
    # (2, 1, 3, 0),
    # (2, 3, 0, 1),
    # (2, 3, 1, 0),
    # (3, 0, 1, 2),
    # (3, 0, 2, 1),
    (3, 1, 0, 2),
    # (3, 1, 2, 0),
    # (3, 2, 0, 1),
    (3, 2, 1, 0),
]

def create_standard_layout(tensor):
    """
    Standard contiguous layout (C-contiguous):
    - Elements in the innermost dimension are adjacent in memory
    - Row-major order: rows are stored contiguously
    - Default PyTorch layout with strides that decrease from left to right
    - Fastest access pattern when iterating through the last dimension first
    """
    return tensor.clone().contiguous()

def create_strided_layout(tensor):
    """
    Strided layout with gaps:
    - Contains same data but with deliberate gaps in memory
    - Non-standard strides that don't match either C or F contiguous patterns
    - Tests kernel's ability to handle arbitrary stride patterns
    - Good for detecting stride-related bugs in kernels assuming contiguity
    """
    # Handle 1D tensors specially
    if tensor.dim() == 1:
        size = tensor.size(0)
        # Create tensor with gaps
        padded = torch.zeros(size * 2, dtype=tensor.dtype, device=tensor.device)
        padded[::2] = tensor  # Place values at even indices
        return padded[::2]    # Return view with same values but different stride

    # For 2D+ tensors
    shape = list(tensor.shape)
    padded_shape = shape.copy()
    padded_shape[1] = shape[1] * 2  # Double size in dimension 1

    padded = torch.zeros(padded_shape, dtype=tensor.dtype, device=tensor.device)
    padded[:, ::2, ...] = tensor     # Place at even indices of dimension 1
    return padded[:, ::2, ...]       # Return view with same values but different stride

def create_permuted_strides_layout(tensor, permutation):
    """
    Creates a tensor with permuted strides while maintaining the same logical indexing.
    
    Args:
        tensor (torch.Tensor): Input tensor to permute strides for
        permutation (list): List of integers representing the desired stride order.
                          Can include indices >= tensor.dim() which will be ignored.
                          If shorter than tensor.dim(), remaining dimensions will be kept in order.
    
    Returns:
        torch.Tensor: A new tensor with permuted strides but same logical indexing
        
    Example:
        >>> x = torch.randn(2, 3, 4)
        >>> # Permute strides to be [dim2, dim0, dim1]
        >>> y = create_permuted_strides_layout(x, [2, 0, 1])
    """
    # Extend permutation to match tensor dimensions if needed
    if len(permutation) < tensor.dim():
        permutation = list(permutation) + list(range(len(permutation), tensor.dim()))

    # Filter out invalid permutation indices
    valid_perm = [p for p in permutation if p < tensor.dim()]

    # If permutation is empty or all indices are invalid, return original tensor
    if not valid_perm:
        return tensor.clone()

    # First permute the dimensions according to the valid permutation
    permuted = tensor.permute(*valid_perm)

    # Make the permuted tensor contiguous to get the desired stride order
    permuted = permuted.contiguous()

    # Create the inverse permutation to get back to original dimension order
    inverse_perm = [valid_perm.index(i) for i in range(len(valid_perm))]

    # Permute back to original dimension order
    # This creates a tensor with the same shape but different strides
    result = permuted.permute(*inverse_perm)

    return result


def diff(a, b, rtol=None, atol=None,
         assert_close=True, verbose=False, title=None,
         pct_tol=None):
    """ A diff function that helps debug numerical issues
    Args:
        a: torch.Tensor
        b: torch.Tensor
        rtol: float
        atol: float
        assert_close: bool
        verbose: bool
        pct_tol: float
    Returns:
        bool: True if a and b are close, False otherwise
    """
    a = a.to(torch.float32)
    b = b.to(torch.float32)
    if rtol is None: rtol = 1e-3
    if atol is None: atol = 1e-5
    equal = torch.allclose(a, b, rtol=rtol, atol=atol)
    error_max = torch.max(torch.abs(a - b))

    pct_equal = ((((a - b).abs() <= atol + rtol * torch.abs(b)).sum().item() / a.numel()) >= pct_tol) if pct_tol is not None else False

    # Calculate absolute error
    abs_diff = torch.abs(a - b)
    total_elements = a.numel()

    # Calculate relative error where b is non-zero
    b_nonzero = b != 0
    rel_diff = torch.zeros_like(abs_diff)
    rel_diff[b_nonzero] = abs_diff[b_nonzero] / torch.abs(b[b_nonzero])
    
    # Check if pct_tol makes it pass
    if not equal and pct_equal:
        equal = True

    # Only construct and log/print message if there are failures
    if not equal:
        # Construct the message
        msg_lines = []
        msg_lines.append('\n')
        msg_lines.append('=' * 10 + f" {title} " + '=' * 10)
        msg_lines.append(f"Max absolute error: {error_max.item()}")
        msg_lines.append(f"Tensors are different according to torch.allclose")

        # Calculate thresholds for relative error table
        rel_thresholds = torch.logspace(
         (torch.log10(torch.tensor(rtol))), 
            0.0, 
            steps=10
        )

        # Calculate thresholds for absolute error table
        abs_thresholds = torch.logspace(
         (torch.log10(torch.tensor(atol))), 
            0.0, 
            steps=10
        )

        # Build relative error table
        msg_lines.append("\nRelative Error Table:")
        msg_lines.append("---------------------")
        msg_lines.append(f"{'Threshold':<12} {'% matched':<12} {'Element Count':<12}")
        msg_lines.append("-" * 36)
        for threshold in rel_thresholds:
            count = (rel_diff <= threshold).sum().item()
            percentage = 100.0 * count / total_elements
            msg_lines.append(f"{threshold.item():<12.6f} {percentage:<12.2f} {count:<12}")

        # Build absolute error table
        msg_lines.append("\nAbsolute Error Table:")
        msg_lines.append("---------------------")
        msg_lines.append(f"{'Threshold':<12} {'% matched':<12} {'Element Count':<12}")
        msg_lines.append("-" * 36)
        for threshold in abs_thresholds:
            count = (abs_diff <= threshold).sum().item()
            percentage = 100.0 * count / total_elements
            msg_lines.append(f"{threshold.item():<12.6f} {percentage:<12.2f} {count:<12}")

        # Build largest errors section - only for points that actually fail tolerance
        # Find points that fail the tolerance check
        tolerance_check = abs_diff <= atol + rtol * torch.abs(b)
        failing_points = ~tolerance_check
        
        if failing_points.any():
            failing_count = int(failing_points.sum().item())
            n_samples = min(5, failing_count)
            msg_lines.append(f"\nLargest Errors (from {failing_count} failing points):")
            
            # Get indices of failing points, sorted by error magnitude
            failing_flat_indices = torch.nonzero(failing_points.flatten(), as_tuple=False).squeeze(1)
            failing_abs_diff = abs_diff.flatten()[failing_flat_indices]
            sort_order = torch.argsort(failing_abs_diff, descending=True)
            sorted_failing_indices = failing_flat_indices[sort_order]
            
            for i in range(n_samples):
                idx = sorted_failing_indices[i]
                multi_idx = torch.unravel_index(idx, a.shape)
                multi_idx_str = ', '.join(map(str, [idx.item() for idx in multi_idx]))
                msg_lines.append(f"Index [{multi_idx_str}]: a={a[multi_idx].item()}, b={b[multi_idx].item()}, "
                      f"abs_diff={abs_diff[multi_idx].item()}, rel_diff={rel_diff[multi_idx].item()}")
        else:
            msg_lines.append("\nNo individual points fail tolerance check (likely due to pct_tol)")

        # Join all lines into a single message
        msg = '\n'.join(msg_lines)
        
        # Log the message
        logger.info(msg)
        
        # Print if verbose is True
        if verbose:
            print(msg)

    if assert_close:
        assert equal, f"Tensors are not close! Max absolute error: {error_max.item()}"

    return equal



def test_permuted_strides_layout():
    """
    Test suite for create_permuted_strides_layout function.
    Verifies both mathematical properties and stride permutations.
    """

    def verify_strides(tensor, expected_perm):
        """Helper function to verify strides match expected relative order permutation"""
        strides = tensor.stride()
        # Get the relative order of strides
        stride_order = sorted(range(len(strides)), key=lambda i: strides[i], reverse=True)
        # Compare with expected permutation (only up to tensor dimensions)
        valid_perm = [p for p in expected_perm if p < tensor.dim()]

        return stride_order == valid_perm

    # Test 1: Basic 2D tensor permutation
    x = torch.randn(3, 4)
    y = create_permuted_strides_layout(x, [1, 0])
    assert y.shape == x.shape
    assert torch.allclose(y, x)  # Mathematical equivalence
    assert verify_strides(y, [1, 0])

    # Test 2: 3D tensor with partial permutation
    x = torch.randn(2, 3, 4)
    y = create_permuted_strides_layout(x, [2, 0, 1])
    assert y.shape == x.shape
    assert torch.allclose(y, x)
    assert verify_strides(y, [2, 0, 1])

    # Test 3: Invalid permutation indices
    x = torch.randn(2, 3)
    y = create_permuted_strides_layout(x, [0, 1, 2, 3])
    assert y.shape == x.shape
    assert torch.allclose(y, x)
    assert verify_strides(y, [0, 1])  # Only valid indices should be considered

    # Test 4: Empty permutation
    x = torch.randn(2, 3)
    y = create_permuted_strides_layout(x, [])
    assert y.shape == x.shape
    assert torch.allclose(y, x)

    # Test 5: 1D tensor
    x = torch.randn(5)
    y = create_permuted_strides_layout(x, [0])
    assert y.shape == x.shape
    assert torch.allclose(y, x)

    # Test 6: Mathematical operations
    x = torch.randn(2, 3, 4)
    y = create_permuted_strides_layout(x, [2, 0, 1])
    z = create_permuted_strides_layout(x, [1, 2, 0])

    # Test addition
    assert torch.allclose(x + y, x + x)
    assert torch.allclose(y + z, x + x)

    # Test multiplication
    assert torch.allclose(x * y, x * x)
    assert torch.allclose(y * z, x * x)

    # Test matrix multiplication (if applicable)
    if x.dim() == 2:
        assert torch.allclose(x @ y, x @ x)

    # Test 7: Different data types
    # For floating point types, use randn
    for dtype in [torch.float32, torch.float64]:
        x = torch.randn(2, 3, dtype=dtype)
        y = create_permuted_strides_layout(x, [1, 0])
        assert y.dtype == dtype
        assert torch.allclose(y, x)
        assert verify_strides(y, [1, 0])

    # For integer types, use randint
    for dtype in [torch.int32, torch.int64]:
        x = torch.randint(0, 10, (2, 3), dtype=dtype)
        y = create_permuted_strides_layout(x, [1, 0])
        assert y.dtype == dtype
        assert torch.allclose(y, x)
        assert verify_strides(y, [1, 0])

    # Test 8: Different devices
    if torch.cuda.is_available():
        x = torch.randn(2, 3, device='cuda')
        y = create_permuted_strides_layout(x, [1, 0])
        assert y.device == x.device
        assert torch.allclose(y, x)
        assert verify_strides(y, [1, 0])

def tensor_exact_copy(tensor):
    """ Create a copy of a tensor that preserves shape, stride, dtype device, and requires_grad """
    shape = tensor.shape
    stride = tensor.stride()
    # Since the strides are arbitrary we might need more (or less) data than shape would imply
    data_size = 0
    for shp, strd in zip(shape, stride):
        data_size += shp * strd
    tensor_copy = torch.empty(data_size, dtype=tensor.dtype, device=tensor.device, requires_grad=False)
    tensor_copy = tensor_copy.as_strided(shape, stride)
    tensor_copy.copy_(tensor)

    tensor_copy.requires_grad_(tensor.requires_grad)
    return tensor_copy

def exact_copy(obj):
    if isinstance(obj, torch.Tensor):
        return tensor_exact_copy(obj)
    elif isinstance(obj, dict):
        return {k: exact_copy(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [exact_copy(v) for v in obj]
    elif isinstance(obj, tuple):
        return tuple(exact_copy(v) for v in obj)
    elif isinstance(obj, (int, float, bool, str)):
        return obj
    raise ValueError(f"Unsupported type: {type(obj)}")

if __name__ == "__main__":
    test_permuted_strides_layout()
    logger.info("All tests passed!")


import os
import yaml
import logging
from typing import Callable

logger = logging.getLogger(__name__)
DEFAULT_TIMING_ROOT = os.path.join(os.path.dirname(os.path.abspath(__file__)), '.timing_cache')


class ConfigTimingCache:
    """A disk-backed cache for storing config timing results.
    
    The cache stores entries of the form:
    {
        hash_key: [
            {
                "timings": [      # List sorted by time (ascending)
                    {
                        "config": {...}, # Configuration that was tested
                        "time": float,   # Timing result
                        "std": float,    # Standard deviation of the timing result
                        "n": int,        # Number of runs
                    },
                    ...
                ]
            },
            ...  # Multiple arguments can hash to the same key
        ]
    }
    
    The cache itself is kept sorted by the best (lowest) timing of each entry,
    allowing fast access to optimal configurations.
    """
    
    def __init__(
        self, 
        fn: str, 
        hash_fn: Callable[[dict], str],
        root: str | None = None, 
        size: int | None = None,
        strict: bool = False
    ):
        """ Args:
                fn: Name of the function being cached
                hash_fn: Function that converts an arguments dict to a string hash
                root: Root directory for cache files. Defaults to .timing_cache/
                size: Maximum number of entries in the cache. Defaults to None (unlimited)
                strict: If True, verify cache is properly sorted on load
        """
        self.root = root or DEFAULT_TIMING_ROOT
        self.location = os.path.join(self.root, f"{fn}.yaml")
        self.fn = fn
        self.size = size
        self.hash_fn = hash_fn
        self._cache: dict[str, list[dict]] = {}
        
        # Create directory if it doesn't exist
        os.makedirs(os.path.dirname(self.location), exist_ok=True)
        
        # Load existing cache if it exists
        if os.path.exists(self.location):
            with open(self.location, 'r') as f:
                loaded_data = yaml.load(f, yaml.Loader)
                if loaded_data is not None:
                    self._cache = loaded_data
                else:
                    self._cache = {}
        else:
            self._cache = {}

        if strict:
            self._verify_sorting()

    def _verify_sorting(self):
        """Verify that cache entries and their timings are properly sorted."""
        # Check that each timings list is sorted
        for k, entries in self._cache.items():
            for entry in entries:
                timings = entry["timings"]
                sorted_timings = sorted(timings, key=lambda x: x["time"])
                if timings != sorted_timings:
                    raise ValueError(f"Timings list not sorted for key {k}")
            entry_best_times = [entry["timings"][0]["time"] for entry in entries]
            if not all(entry_best_times[i] <= entry_best_times[i+1] for i in range(len(entry_best_times) - 1)):
                raise ValueError("Cache entries not sorted by best timing")
    
    def __getitem__(self, arguments: dict) -> dict:
        """Get the best performing config for the given input arguments.
        
        Args:
            arguments: Input arguments to look up
            
        Returns:
            The config with the lowest timing for this input arguments
        """
        key = self.hash_fn(arguments)
        return self._cache[key][0]["timings"][0]["config"]
    
    def get_timings(self, arguments: dict) -> list[dict]:
        """Get the timings for the given input arguments."""
        key = self.hash_fn(arguments)
        return self._cache[key][0]["timings"]
    
    def __contains__(self, arguments: dict) -> bool:
        key = self.hash_fn(arguments)
        if key not in self._cache:
            logger.debug(f"Cache miss for {self.fn} with {key=}")
            return False
        logger.debug(f"Cache hit for {self.fn} with {key=}")
        return True

    def __len__(self) -> int:
        return len(self._cache)

    def update(self, arguments: dict, timings: list[dict]):
        """Update cache with new timings for arguments.
        
        Args:
            arguments: The input arguments
            timings: List of dicts, each containing:
                    - config: Dict of configuration parameters
                    - time: Float timing result
                    
        The entry will be inserted into the cache maintaining the sort order
        by best timing. The timings list itself will also be sorted by time.
        """
        key = self.hash_fn(arguments)
        
        # Sort timings by time
        sorted_timings = sorted(timings, key=lambda x: x["time"])
        
        entry = {"timings": sorted_timings}
        
        if key in self._cache:
            self._cache[key].append(entry)
        else:
            self._cache[key] = [entry]

        self._cache[key] = sorted(self._cache[key], key=lambda x: x["timings"][0]["time"])
        
        # Enforce size limit if needed
        while self.size is not None and len(self) > self.size:
            self._cache.popitem()
            
        # Save to disk, ensuring sorted order is preserved
        try:
            with open(self.location, 'w') as f:
                yaml.dump(self._cache, f)
        except Exception as e:
            logger.warning(f"Failed to save cache to {self.location}: {e}")

    def remove(self):
        """Remove the cache file from disk."""
        if self.location is None:
            raise ValueError("Cache location is not set")
        os.remove(self.location)

    def store(self, path: str):
        """Store the cache to a file."""
        with open(path, 'w') as f:
            yaml.dump(self._cache, f)

    def to_df(self):
        """Convert the cache to a pandas DataFrame."""
        import pandas as pd
        return pd.DataFrame(self._cache)


import hashlib
import functools
import os
import re
import subprocess
import uuid
import torch
import threading
from functools import lru_cache
from torch.utils.cpp_extension import CUDA_HOME
from pathlib import Path
from typing import Tuple, Iterable, Optional, Any, Dict
import tempfile
from vidrial.jit.runtime import Runtime
import logging
import shlex

logger = logging.getLogger(__name__)


_cutlass_lock = threading.Lock()


def hash_to_hex(s: str) -> str:
    md5 = hashlib.md5()
    md5.update(s.encode('utf-8'))
    return md5.hexdigest()[0:12]


@functools.lru_cache(maxsize=None)
def get_nvcc_compiler() -> Tuple[str, str]:
    paths = []
    if os.getenv('NVCC_COMPILER'):
        paths.append(os.getenv('NVCC_COMPILER'))
    paths.append(f'{CUDA_HOME}/bin/nvcc')

    # Try to find the first available NVCC compiler
    least_version_required = '12.2'
    version_pattern = re.compile(r'release (\d+\.\d+)')
    for path in paths:
        if os.path.exists(path):
            match = version_pattern.search(os.popen(f'{path} --version').read())
            version = match.group(1) # type: ignore
            assert match, f'Cannot get the version of NVCC compiler {path}'
            assert version >= least_version_required, f'NVCC {path} version {version} is lower than {least_version_required}'
            return path, version
    raise RuntimeError('Cannot find any available NVCC compiler')

def get_cuda_arch() -> Optional[str]:
    """Detect CUDA architecture of the local GPU, return None if CUDA is not available"""
    try:
        if not torch.cuda.is_available():
            return None
        dprops = torch.cuda.get_device_properties(torch.device('cuda'))
        name, major, minor = dprops.name, dprops.major, dprops.minor
        arch = f'{major}{minor}'
        arch = '90a' if arch == '90' else arch
        return arch
    except Exception:
        logger.warning('Cannot detect CUDA architecture, assuming 86')
        return None

def vidrial_path():
    return Path(__file__).parent.parent

def vidrial_include_dirs() -> list[str]:
    root = vidrial_path()
    kernels = root / 'kernels'
    cuda_utils = root / 'cuda_utils'
    return [str(p) for p in [root, kernels, cuda_utils]]

def ensure_cutlass(path: str):
    """ Checkout cutlass repo into path
    """
    subprocess.run(['git', 'clone', 'https://github.com/m-a-n-i-f-e-s-t/cutlass.git', path], check=False)

def cutlass_include_dirs() -> list[str]:
    cutlass_dir = Path(__file__).parent.parent.parent / 'cutlass'
    if not cutlass_dir.exists(): # In a user environment
        cutlass_dir = Path(__file__).parent.parent / 'cutlass'
        with _cutlass_lock:
            if not cutlass_dir.exists():
                git_dir = cutlass_dir / '.git'
                if not git_dir.exists():
                    cutlass_dir.parent.mkdir(parents=True, exist_ok=True)
                    ensure_cutlass(str(cutlass_dir))
                    if not git_dir.exists():
                        raise RuntimeError(f"Failed to clone cutlass repository to {cutlass_dir}")
    
    cutlass_include = cutlass_dir / 'include'
    return [str(cutlass_include)]

@lru_cache(maxsize=None)
def get_kernel_hash(code: str) -> str:
    """ A proper hash of the kernel requires a proper build system that tracks dependencies.
    This is a lightweight hash function that assumes that the JITted code only relies on vidrial headers (and cutlass).
    """
    md5 = hashlib.md5()
    for root, _, files in os.walk(str(vidrial_path())):
        for filename in sorted(files):
            if filename.endswith('.cuh'):
                with open(os.path.join(root, filename), 'rb') as f:
                    md5.update(f.read())

    md5.update(code.encode('utf-8'))
    return md5.hexdigest()[0:12]


DEFAULT_INCLUDE_DIRS: tuple[str] = tuple(vidrial_include_dirs() + cutlass_include_dirs()) # type: ignore

def put(path, data, is_binary=False):
    # Write and do POSIX atomic replace
    tmpdir = tempfile.mkdtemp()
    tmp_file_path = f'{tmpdir}/file.tmp.{str(uuid.uuid4())}.{hash_to_hex(path)}'
    with open(tmp_file_path, 'wb' if is_binary else 'w') as f:
        f.write(data)
    os.replace(tmp_file_path, path)


def _make_hashable(x: Any) -> Any:
    """Convert a value into a hashable form for caching."""
    if isinstance(x, torch.Tensor):
        # For tensors, we use shape, dtype, and device as the key
        return (tuple(x.shape), str(x.dtype), str(x.device))
    elif isinstance(x, dict):
        return tuple((k, _make_hashable(v)) for k, v in sorted(x.items()))
    elif isinstance(x, (list, tuple)):
        return tuple(_make_hashable(v) for v in x)
    elif isinstance(x, (int, float, str, bool, type(None))):
        return x
    elif hasattr(x, '__str__'):
        # Handle C++ type objects that can be stringified
        return str(x)
    else:
        raise TypeError(f"Cannot make {type(x)} hashable for caching")


def compiler_flags(arch: Optional[str] = None) -> list[str]:
    if arch is None:
        arch = get_cuda_arch() or '86'
    nvcc_flags = ['-std=c++20', '-w', '-shared', '-O3', '-lineinfo', '--expt-relaxed-constexpr', '--expt-extended-lambda', '--use_fast_math','-ftemplate-backtrace-limit=0', '--resource-usage',
                  f'-gencode=arch=compute_{arch},code=sm_{arch}',
                  '--ptxas-options=--register-usage-level=10' + (',--verbose' if 'PTXAS_VERBOSE' in os.environ else ''),
                  # Suppress some unnecessary warnings, such as unused variables for certain `constexpr` branch cases
                  '--diag-suppress=177,174,940']
    cxx_flags = ['-fPIC', '-O3', '-Wno-deprecated-declarations', '-Wno-abi']
    
    if os.getenv('GCC_PROFILE_LEVEL', '0') == '1':  # Basic profiling
        cxx_flags.extend(['-ftime-report', '-fmem-report'])

    flags = [*nvcc_flags, f'--compiler-options={",".join(cxx_flags)}']
    return flags


def compile(code: str, include_dirs: Optional[tuple] = None, arch: Optional[str] = None, path: Optional[str] = None, flags: Optional[list[str]] = None) -> Runtime:
    """ Compile a CUDA kernel and return a Runtime object.

    Args:
        code: Code of the kernel.
        include_dirs: Include directories.
        arch: CUDA architecture.
        path: Path where shared object will be stored. Default to a temporary directory.

    Returns:
        A Runtime object.
    """
    if include_dirs is None:
        include_dirs = tuple(DEFAULT_INCLUDE_DIRS)

    if flags is None:
        flags = compiler_flags(arch)

    if path is None:
        path = tempfile.mkdtemp()

    # Write the code
    os.makedirs(path, exist_ok=True)
    src_path = f'{path}/kernel.cu'

    put(src_path, code)

    # Compile into a temporary SO file
    so_path = f'{path}/kernel.so'
    tmp_so_path = f'{path}/nvcc.tmp.{str(uuid.uuid4())}.{hash_to_hex(so_path)}.so'

    # Compile
    command = [
        get_nvcc_compiler()[0], src_path,
        '-o', tmp_so_path,
        *flags,
        *[f'-I{d}' for d in include_dirs]
    ]
    # Format command as a copy-pasteable shell command
    command_str = ' '.join(shlex.quote(arg) for arg in command)
    logger.debug(f'Compiling runtime with command:\n{command_str}')

    res = subprocess.run(command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    if res.returncode != 0:
        msg = f'Failed to compile {src_path}, {res.stdout.decode("utf-8") if res.stdout else "No stdout"}'
        logger.error(msg)
        raise RuntimeError(msg)
    
    logger.debug(f'Compilation output: {res.stdout.decode("utf-8") if res.stdout else ""}')

    # Atomic replace SO file
    os.replace(tmp_so_path, so_path)
    return Runtime(path)



"""
Interactive visualization dashboard for CUDA kernel timing analysis.

This module provides functionality to:
- Load timing data from YAML files
- Generate interactive dashboards with filtering capabilities
- Support snapshot comparison for performance analysis
- Create publication-ready visualizations
"""

import yaml
import json
import os
from typing import Dict, List, Any, Set, Union
from pathlib import Path


def custom_yaml_constructor(loader, tag_suffix, node):
    """Handle custom YAML tags by converting them to string representations."""
    if isinstance(node, yaml.ScalarNode):
        return loader.construct_scalar(node)
    elif isinstance(node, yaml.SequenceNode):
        return loader.construct_sequence(node)
    elif isinstance(node, yaml.MappingNode):
        return loader.construct_mapping(node)
    else:
        return str(node)


def load_timing_data(yaml_path: str) -> Dict[str, List[Dict]]:
    """
    Load timing data from YAML file with custom object handling.
    
    Args:
        yaml_path: Path to the YAML timing cache file
        
    Returns:
        Dictionary mapping problem shapes to lists of timing configurations
    """
    # Create a custom YAML loader that handles unknown tags
    CustomLoader = yaml.SafeLoader
    CustomLoader.add_multi_constructor('', custom_yaml_constructor)
    
    with open(yaml_path, 'r') as f:
        data = yaml.load(f, Loader=CustomLoader)
    
    return data


def process_timing_data_for_dashboard(timings_data: Dict[str, Any]) -> Dict[str, Any]:
    """
    Process timing data to match the dashboard.js data structure expectations.
    
    Args:
        timings_data: Raw timing data from YAML
        
    Returns:
        Dictionary with processed data for dashboard
    """
    # Flatten all timing records and add problem_shape field
    raw_data = []
    
    for problem_shape_hash, problem_list in timings_data.items():
        # Each problem shape maps to a list, usually with one item containing 'timings'
        for problem_data in problem_list:
            if 'timings' in problem_data:
                for timing_entry in problem_data['timings']:
                    # Each timing entry has 'config' and timing results
                    if 'config' in timing_entry:
                        record = timing_entry['config'].copy()
                        
                        # Convert complex values to strings for consistent filtering
                        for key, value in record.items():
                            if isinstance(value, (list, dict)):
                                record[key] = str(value)
                        
                        # Add timing results to the record
                        for key, value in timing_entry.items():
                            if key != 'config':
                                record[key] = value
                        
                        # Add problem shape identifier
                        record['problem_shape'] = problem_shape_hash
                        raw_data.append(record)
    
    # Extract configuration attributes from the 'config' sections
    config_attributes = {}
    timing_keys = {'time', 'n', 'std', 'gflops', 'memory_gb', 'tensor_utilization', 'problem_shape'}
    
    for record in raw_data:
        for key, value in record.items():
            if key not in timing_keys:
                if key not in config_attributes:
                    config_attributes[key] = {
                        'values': set(),
                        'type': type(value).__name__
                    }
                
                # Convert unhashable types to strings for set storage
                if isinstance(value, (list, dict)):
                    value_str = str(value)
                    # Also store the original type info for proper filtering
                    config_attributes[key]['has_complex_values'] = True
                else:
                    value_str = value
                    
                config_attributes[key]['values'].add(value_str)
    
    # Convert sets to sorted lists for JSON serialization
    for attr_data in config_attributes.values():
        attr_data['values'] = sorted(list(attr_data['values']))
    
    # Get unique problem shapes
    problem_shapes = sorted(list(timings_data.keys()))
    
    return {
        'rawData': raw_data,
        'configAttributes': config_attributes,
        'problemShapes': problem_shapes
    }


def get_dashboard_assets() -> Dict[str, str]:
    """
    Load dashboard HTML, CSS, and JavaScript assets from the viz directory.
    
    Returns:
        Dictionary containing 'html', 'css', and 'js' asset contents
    """
    viz_dir = Path(__file__).parent / 'viz'
    
    assets = {}
    
    # Load HTML template
    html_path = viz_dir / 'dashboard.html'
    if html_path.exists():
        with open(html_path, 'r') as f:
            assets['html'] = f.read()
    else:
        raise FileNotFoundError(f"HTML template not found at {html_path}")
    
    # Load CSS
    css_path = viz_dir / 'styles.css'
    if css_path.exists():
        with open(css_path, 'r') as f:
            assets['css'] = f.read()
    else:
        raise FileNotFoundError(f"CSS file not found at {css_path}")
    
    # Load JavaScript
    js_path = viz_dir / 'dashboard.js'
    if js_path.exists():
        with open(js_path, 'r') as f:
            assets['js'] = f.read()
    else:
        raise FileNotFoundError(f"JavaScript file not found at {js_path}")
    
    return assets


def generate_dashboard_html(timings_data: Dict[str, List[Dict]], output_path: str) -> None:
    """
    Generate complete HTML dashboard with embedded assets and data.
    
    Args:
        timings_data: Dictionary of problem shapes to timing configurations
        output_path: Path where the HTML file should be saved
    """
    # Process the data for the dashboard
    dashboard_data = process_timing_data_for_dashboard(timings_data)
    
    # Load dashboard assets
    assets = get_dashboard_assets()
    
    # Build the complete HTML
    html_template = assets['html']
    
    # Replace external CSS link with inline styles
    html_with_css = html_template.replace(
        '<link rel="stylesheet" href="styles.css">',
        f'<style>\n{assets["css"]}\n</style>'
    )
    
    # Replace the dashboard data placeholder
    html_with_data = html_with_css.replace(
        'window.DASHBOARD_DATA = {\n            rawData: [],\n            configAttributes: {},\n            problemShapes: []\n        };',
        f'window.DASHBOARD_DATA = {json.dumps(dashboard_data, indent=8)};'
    )
    
    # Replace external JS script with inline script
    final_html = html_with_data.replace(
        '<script src="dashboard.js"></script>',
        f'<script>\n{assets["js"]}\n</script>'
    )
    
    # Write the final HTML file
    with open(output_path, 'w') as f:
        f.write(final_html)
    
    print(f"Dashboard generated: {output_path}")
    print(f"Data summary: {len(dashboard_data['problemShapes'])} problem shapes, {len(dashboard_data['rawData'])} total configurations")


def main():
    """Main function for command-line usage."""
    import sys
    
    if len(sys.argv) < 2:
        print("Usage: python viz.py <yaml_file> [output_file]")
        return
    
    yaml_file = sys.argv[1]
    output_file = sys.argv[2] if len(sys.argv) > 2 else "kernel_dashboard.html"
    
    try:
        data = load_timing_data(yaml_file)
        generate_dashboard_html(data, output_file)
        print(f"Dashboard successfully generated: {output_file}")
    except Exception as e:
        print(f"Error generating dashboard: {e}")


if __name__ == "__main__":
    main()


import torch
import pytest
from vidrial.jit.ctypes import dtype_map
from vidrial.jit.jit import jit

@pytest.fixture
def get_add_one_kernel_code():
    def _get_add_one_kernel_code(input: torch.Tensor, output: torch.Tensor, tile_d0: int = 32, tile_d1: int = 32, thread_num: int = 32):
        d0, d1 = input.shape
        assert input.dtype == output.dtype, f"Input and output must have the same dtype, got {input.dtype} and {output.dtype}"
        return """
#include "add_one/add_one_kernels.cuh"
#include "copy/copy_kernels.cuh"
using namespace cute;
using namespace vidrial;

extern "C" {{
    void launch(float* In, float* Out) {{
        using SlabShape = Shape<Int<{d0}>, Int<{d1}>>;
        using TileShape = Shape<Int<{tile_d0}>, Int<{tile_d1}>>;
        using ASlab = Layout<SlabShape>;
        using T = {T};
        //T* in_ptr = reinterpret_cast<T*>(In);
        //T* out_ptr = reinterpret_cast<T*>(Out);
        launch_add_one<float, {thread_num}, SlabShape, TileShape, ASlab>(In, Out);
    }}
}}
        """.format(d0=d0, d1=d1, tile_d0=tile_d0, tile_d1=tile_d1, thread_num=thread_num, T=dtype_map[input.dtype])
    return _get_add_one_kernel_code


@pytest.fixture
def add_one_kernel(get_add_one_kernel_code, tmp_path):
    def _add_one_kernel(X, Y, d0_tile, d1_tile, thread_num):
        code = get_add_one_kernel_code(X, Y, d0_tile, d1_tile, thread_num)
        runtime = jit("test_add_one_kernel", code, root=str(tmp_path))
        runtime(X, Y)
        return Y
    return _add_one_kernel


import torch as th
from math import comb, factorial, sqrt, prod

def tprod(*xs) -> th.Tensor:
    batch_dims = tuple(xs[0].shape[:-1])
    for xi in xs:
        assert xi.shape[:-1] == batch_dims, f"batch_dims of xi={xi.shape[:-1]} do not match batch_dims={batch_dims}"
    y_shape = batch_dims + tuple([xi.shape[-1] for xi in xs])
    y = th.ones(y_shape, device=xs[0].device, dtype=xs[0].dtype)
    batch_slice = [slice(None)] * len(batch_dims)
    for i in range(len(xs)):
        feature_bcast: List[Union[None, slice]] = [None] * len(xs)
        feature_bcast[i] = slice(None)
        y *= xs[-i-1][batch_slice + feature_bcast]
    return y

def tpow(x, power: int) -> th.Tensor:
    return tprod(*([x]*power))

def tprod_bcast(x, power, dim) -> th.Tensor:
    batch_dims = tuple(x.shape[:-1])
    batch_slice = [slice(None)] * len(batch_dims)
    feature_bcast: List[Union[None, slice]] = [None] * power
    feature_bcast[dim] = slice(None)
    x_bcast = x[batch_slice + feature_bcast]
    return x_bcast

def tprod_bwd(xis, zgrad, power):
    b_dims = len(zgrad.shape) - power
    xigrads = []
    for i in range(power):
        v = zgrad.clone()
        for j in range(power):
            if j != i: v = v * tprod_bcast(xis[-j-1], power, j)
        reduce_dims = [j+b_dims for j in range(power) if j != i]
        xigrad_delta = th.sum(v, dim=tuple(reduce_dims))
        xigrads.append(xigrad_delta)
    return xigrads

def tpow_shape(x_shape, power):
    return th.Size(tuple(x_shape[:-1]) + (x_shape[-1],) * power)

def default_d_tile(d, power):
    """
    Returns the default d_tile for a given d and power.
    """
    default_d_tiles = {
        16: {
            2: 4,
            3: 4,
            4: 2,
        },
        32: {
            2: 8,
            3: 4,
            4: 2
        },
        64: {
            2: 8,
            3: 4,
            4: 2
        }
    }
    try:
        return default_d_tiles[d][power]
    except KeyError:
        return 1
